#-----------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import sys
from datetime import datetime
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from textwrap import dedent
from unittest import skipIf

from django.test import TestCase
from django.core.exceptions import ValidationError
from django.contrib.gis.geos import GEOSGeometry, Polygon, MultiPolygon
from django.utils.dateparse import parse_datetime
from django.utils.timezone import utc

from eoxserver.core import env
from eoxserver.resources.coverages.models import *
from eoxserver.resources.coverages.util import collect_eo_metadata
from eoxserver.resources.coverages.metadata.coverage_formats import (
    native, eoom, dimap_general
)


def create(Class, **kwargs):
    obj = Class(**kwargs)
    obj.full_clean()
    obj.save()
    return obj


def refresh(*objects):
    refr = lambda obj: type(obj).objects.get(identifier=obj.identifier)
    if len(objects) == 1:
        return refr(objects[0])
    return map(refr, objects)


def union(*footprints):
    u = None
    for footprint in footprints:
        if u is None:
            u = footprint
        else:
            u = footprint.union(u)
    return u


class GeometryMixIn(object):
    def assertGeometryEqual(self, ga, gb, tolerance=0.05, max_area=0.00001):
        try:
            if ga.equals(gb):
                return
            if ga.difference(gb).empty:
                return
            if ga.equals_exact(gb, tolerance):
                return
            if ga.difference(gb).area < max_area:
                return
        except Exception:
            pass

        self.fail(
            ga.equals_exact(gb, tolerance),
            "%r != %r" % (ga.wkt, gb.wkt)
        )

class ModelTests(GeometryMixIn, TestCase):
    def setUp(self):
        self.coverage_type = create(CoverageType,
            name="RGB"
        )
        self.rectified_grid = Grid.objects.create(
          name="rectified_grid",
          coordinate_reference_system="EPSG:4326",
          axis_1_name="x",
          axis_1_type=0,
          axis_1_offset="5",
          axis_2_name="y",
          axis_2_type=0,
          axis_2_offset="5"
        )
        self.referenced_grid= Grid.objects.create(
          name="referenced_grid",
          coordinate_reference_system="EPSG:4326",
          axis_1_name="x",
          axis_1_type=0,
          axis_2_name="y",
          axis_2_type=0

        )
        self.rectified_1 = create(Coverage,
            identifier="rectified-1",
            footprint=GEOSGeometry("MULTIPOLYGON (((-111.6210939999999994 26.8588260000000005, -113.0273439999999994 -4.0786740000000004, -80.6835939999999994 -9.7036739999999995, -68.0273439999999994 15.6088260000000005, -111.6210939999999994 26.8588260000000005)))"),
            begin_time=parse_datetime("2013-06-11T14:55:23Z"),
            end_time=parse_datetime("2013-06-11T14:55:23Z"),
            grid=self.rectified_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=self.coverage_type
        )

        self.rectified_2 = create(Coverage,
            identifier="rectified-2",
            footprint=GEOSGeometry("MULTIPOLYGON (((-28.0371090000000009 19.4760129999999982, -32.9589840000000009 -0.9146120000000000, -2.8125000000000000 -3.8150019999999998, 4.2187500000000000 19.1244510000000005, -28.0371090000000009 19.4760129999999982)))"),
            begin_time=parse_datetime("2013-06-10T18:52:34Z"),
            end_time=parse_datetime("2013-06-10T18:52:32Z"),
            grid=self.rectified_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=self.coverage_type
        )

        self.rectified_3 = create(Coverage,
            identifier="rectified-3",
            footprint=GEOSGeometry("MULTIPOLYGON (((-85.5175780000000003 14.2904660000000003, -116.2792969999999997 -8.3853150000000003, -63.7207030000000003 -19.4595340000000014, -58.7988280000000003 7.2592160000000003, -85.5175780000000003 14.2904660000000003)))"),
            begin_time=parse_datetime("2013-06-10T18:55:54Z"),
            end_time=parse_datetime("2013-06-10T18:55:54Z"),
            grid=self.rectified_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=self.coverage_type
        )

        self.referenceable = create(Coverage,
            identifier="referenceable-1",
            footprint=GEOSGeometry("MULTIPOLYGON (((-85.5175780000000003 14.2904660000000003, -116.2792969999999997 -8.3853150000000003, -63.7207030000000003 -19.4595340000000014, -58.7988280000000003 7.2592160000000003, -85.5175780000000003 14.2904660000000003)))"),
            begin_time=parse_datetime("2013-06-10T18:55:54Z"),
            end_time=parse_datetime("2013-06-10T18:55:54Z"),
            grid=self.referenced_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=self.coverage_type
        )

        self.mosaic = create(Mosaic,
            identifier="mosaic-1",
            grid=self.rectified_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=self.coverage_type
        )

        # TODO: bug, requires identifier to be set manually again
        self.mosaic.identifier = "mosaic-1"
        self.mosaic.full_clean()
        self.mosaic.save()

        #=======================================================================
        # Collections
        #=======================================================================

        self.series_1 = create(Collection,
            identifier="series-1"
        )

        self.series_2 = create(Collection,
            identifier="series-2"
        )

    def tearDown(self):
        pass

    def test_insertion(self):
        rectified_1, rectified_2, rectified_3 = (
            self.rectified_1, self.rectified_2, self.rectified_3
        )
        mosaic, series_1, series_2 = self.mosaic, self.series_1, self.series_2

        mosaic_insert_coverage(mosaic, rectified_1)
        mosaic_insert_coverage(mosaic, rectified_2)
        mosaic_insert_coverage(mosaic, rectified_3)
        mosaic_list = mosaic.coverages.all()

        self.assertIn(rectified_1, mosaic_list)
        self.assertIn(rectified_2, mosaic_list)
        self.assertIn(rectified_3, mosaic_list)


        collection_insert_eo_object(series_1, rectified_1)
        collection_insert_eo_object(series_1, rectified_2)
        collection_insert_eo_object(series_1, rectified_3)

        series_1_list= series_1.coverages.all()

        self.assertIn(rectified_1, series_1_list)
        self.assertIn(rectified_2, series_1_list)
        self.assertIn(rectified_3, series_1_list)

        collection_insert_eo_object(series_2, rectified_1)
        collection_insert_eo_object(series_2, rectified_2)
        collection_insert_eo_object(series_2, rectified_3)

        series_2_list = series_2.coverages.all()

        self.assertIn(rectified_1, series_2_list)
        self.assertIn(rectified_2, series_2_list)
        self.assertIn(rectified_3, series_2_list)

        self.assertEqual(len(mosaic_list), 3)
        self.assertEqual(len(series_1_list), 3)
        self.assertEqual(len(series_2_list), 3)

        mosaic = Mosaic.objects.get(identifier="mosaic-1")
        mosaic, series_1, series_2 = refresh(mosaic, series_1, series_2)

        # TODO: further check metadata
        self.assertTrue(series_1.begin_time is not None)

        begin_time, end_time, all_rectified_footprints = collect_eo_metadata(
            Coverage.objects.all()
        )
        time_extent = begin_time, end_time

        extent_footprint = MultiPolygon(
            Polygon.from_bbox(all_rectified_footprints.extent)
        )

        self.assertGeometryEqual(series_1.footprint, all_rectified_footprints)
        self.assertGeometryEqual(series_2.footprint, all_rectified_footprints)
        self.assertGeometryEqual(mosaic.footprint, all_rectified_footprints)

        series_1_time_extent = series_1.begin_time, series_1.end_time
        series_2_time_extent = series_2.begin_time, series_2.end_time
        mosaic_time_extent = mosaic.begin_time, mosaic.end_time

        self.assertEqual(series_1_time_extent, time_extent)
        self.assertEqual(series_2_time_extent, time_extent)
        self.assertEqual(mosaic_time_extent, time_extent)

        for eo_obj in series_1_list:
            pass

    def test_insertion_cascaded(self):
        rectified_1, mosaic, series_1, series_2 = (
            self.rectified_1, self.mosaic, self.series_1, self.series_2
        )
        mosaic_insert_coverage(mosaic, rectified_1)
        collection_insert_eo_object(series_1, rectified_1)
        collection_insert_eo_object(series_2, rectified_1)

        series_1 = refresh(series_1)

        series_1_list= series_1.coverages.all()
        series_2_list= series_2.coverages.all()
        mosaic_list= mosaic.coverages.all()

        self.assertIn(rectified_1, series_2_list)
        self.assertIn(rectified_1, series_1_list)

        self.assertIn(rectified_1, mosaic_list)

        # for obj in series_2.iter_cast(True):
        #     pass

    def test_insertion_failed(self):
        referenceable, mosaic = self.referenceable, self.mosaic

        with self.assertRaises(ManagementError):
            mosaic_insert_coverage(mosaic, referenceable)

        mosaic = refresh(mosaic)
        mosaic_list= mosaic.coverages.all()
        self.assertNotIn(referenceable, mosaic_list)

    def test_insertion_and_removal(self):
        rectified_1, rectified_2, series_1 = (
            self.rectified_1, self.rectified_2, self.series_1
        )
        collection_insert_eo_object(series_1, rectified_1)
        collection_insert_eo_object(series_1, rectified_2)

        series_1 = refresh(series_1)


        collection_exclude_eo_object(series_1, rectified_2)

        series_1 = refresh(series_1)

        rectified_1_time_extent = rectified_1.begin_time, rectified_1.end_time
        series_1_time_extent = series_1.begin_time, series_1.end_time
        self.assertEqual(rectified_1_time_extent, series_1_time_extent)
        self.assertGeometryEqual(rectified_1.footprint,series_1.footprint)

    def test_propagate_eo_metadata_change(self):
        rectified_1, series_1 = self.rectified_1, self.series_1


        new_begin_time = parse_datetime("2010-06-11T14:55:23Z")
        new_end_time = parse_datetime("2010-06-11T14:55:23Z")

        rectified_1.begin_time = new_begin_time
        rectified_1.end_time = new_end_time
        rectified_1.full_clean()
        rectified_1.save()
        collection_insert_eo_object(series_1, rectified_1)

        series_1 = refresh(series_1)

        self.assertEqual(series_1.begin_time, new_begin_time)
        self.assertEqual(series_1.end_time, new_end_time)



class MetadataFormatTests(GeometryMixIn, TestCase):
    def test_native_reader(self):
        xml = """
        <Metadata>
            <EOID>some_unique_id</EOID>
            <BeginTime>2013-08-27T10:00:00Z</BeginTime>
            <EndTime>2013-08-27T10:00:10Z</EndTime>
            <Footprint>
                <Polygon>
                    <Exterior>0 0 20 0 20 10 0 10 0 0</Exterior>
                    <!--<Interior></Interior>-->
                </Polygon>
                <Polygon>
                    <Exterior>10 10 40 10 40 30 10 30 10 10</Exterior>
                    <!--<Interior></Interior>-->
                </Polygon>

            </Footprint>
        </Metadata>
        """
        reader = native.NativeFormat(env)
        self.assertTrue(reader.test(xml))
        values = reader.read(xml)

        self.assertEqual({
            "identifier": "some_unique_id",
            "begin_time": datetime(2013, 8, 27, 10, 0, 0, tzinfo=utc),
            "end_time": datetime(2013, 8, 27, 10, 0, 10, tzinfo=utc),
            "footprint": MultiPolygon(
                Polygon.from_bbox((0, 0, 10, 20)),
                Polygon.from_bbox((10, 10, 30, 40))
            ),
            "format": "native"
        }, values)

    @skipIf(sys.version_info.major == 3, 'segfault in Django')
    def test_native_writer(self):
        values = {
            "identifier": "some_unique_id",
            "begin_time": datetime(2013, 8, 27, 10, 0, 0, tzinfo=utc),
            "end_time": datetime(2013, 8, 27, 10, 0, 10, tzinfo=utc),
            "footprint": MultiPolygon(
                Polygon.from_bbox((0, 0, 10, 20)),
                Polygon.from_bbox((10, 10, 30, 40))
            )
        }
        writer = native.NativeFormat(env)

        f = StringIO()
        writer.write(values, f, pretty=True)
        self.assertEqual(dedent("""\
            <Metadata>
              <EOID>some_unique_id</EOID>
              <BeginTime>2013-08-27T10:00:00Z</BeginTime>
              <EndTime>2013-08-27T10:00:10Z</EndTime>
              <Footprint>
                <Polygon>
                  <Exterior>0.000000 0.000000 20.000000 0.000000 20.000000 10.000000 0.000000 10.000000 0.000000 0.000000</Exterior>
                </Polygon>
                <Polygon>
                  <Exterior>10.000000 10.000000 40.000000 10.000000 40.000000 30.000000 10.000000 30.000000 10.000000 10.000000</Exterior>
                </Polygon>
              </Footprint>
            </Metadata>
            """), dedent(f.getvalue()))

    def test_eoom_reader(self):
        xml = """<?xml version="1.0" encoding="utf-8"?>
        <?xml-stylesheet type="text/xsl" href="schematron_result_for_eop.xsl"?>
        <eop:EarthObservation xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/opt/2.0 ../xsd/opt.xsd" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:gml="http://www.opengis.net/gml/3.2" xmlns:eop="http://www.opengis.net/eop/2.0" xmlns:swe="http://www.opengis.net/swe/1.0" xmlns:om="http://www.opengis.net/om/2.0" gml:id="eop_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
          <om:phenomenonTime>
            <gml:TimePeriod gml:id="tp_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
              <gml:beginPosition>2006-08-16T09:09:29Z</gml:beginPosition>
              <gml:endPosition>2006-08-16T09:12:46Z</gml:endPosition>
            </gml:TimePeriod>
          </om:phenomenonTime>
          <om:resultTime>
            <gml:TimeInstant gml:id="archivingdate_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
              <gml:timePosition>2006-08-16T11:03:08Z</gml:timePosition>
            </gml:TimeInstant>
          </om:resultTime>
          <om:procedure>
            <eop:EarthObservationEquipment gml:id="equ_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
              <eop:platform>
                <eop:Platform>
                  <eop:shortName>ENVISAT</eop:shortName>
                </eop:Platform>
              </eop:platform>
              <eop:instrument>
                <eop:Instrument>
                  <eop:shortName>MERIS</eop:shortName>
                </eop:Instrument>
              </eop:instrument>
              <eop:sensor>
                <eop:Sensor>
                  <eop:sensorType>OPTICAL</eop:sensorType>
                </eop:Sensor>
              </eop:sensor>
            </eop:EarthObservationEquipment>
          </om:procedure>
          <om:observedProperty xlink:href="#params1" />
          <om:featureOfInterest>
            <eop:Footprint gml:id="footprint_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
              <eop:multiExtentOf>
                <gml:MultiSurface gml:id="multisurface_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed" srsName="EPSG:4326">
                  <gml:surfaceMember>
                    <gml:Polygon gml:id="polygon_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
                      <gml:exterior>
                        <gml:LinearRing>
                          <gml:posList>20 10 40 10 40 30 20 30 20 10</gml:posList>
                        </gml:LinearRing>
                      </gml:exterior>
                    </gml:Polygon>
                  </gml:surfaceMember>
                  <gml:surfaceMember>
                    <gml:Polygon gml:id="polygon_MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed">
                      <gml:exterior>
                        <gml:LinearRing>
                          <gml:posList>60 50 80 50 80 70 60 70 60 50</gml:posList>
                        </gml:LinearRing>
                      </gml:exterior>
                    </gml:Polygon>
                  </gml:surfaceMember>
                </gml:MultiSurface>
              </eop:multiExtentOf>
            </eop:Footprint>
          </om:featureOfInterest>
          <om:result />
          <eop:metaDataProperty>
            <eop:EarthObservationMetaData>
              <eop:identifier>MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed</eop:identifier>
              <eop:acquisitionType>NOMINAL</eop:acquisitionType>
              <eop:productType>MER_FRS_1P</eop:productType>
              <eop:status>ARCHIVED</eop:status>
              <eop:downlinkedTo>
                <eop:DownlinkInformation>
                  <eop:acquisitionStation>PDHS-E</eop:acquisitionStation>
                </eop:DownlinkInformation>
              </eop:downlinkedTo>
              <eop:processing>
                <eop:ProcessingInformation>
                  <eop:processingCenter>PDHS-E</eop:processingCenter>
                </eop:ProcessingInformation>
              </eop:processing>
            </eop:EarthObservationMetaData>
          </eop:metaDataProperty>
        </eop:EarthObservation>
        """
        reader = eoom.EOOMFormatReader(env)
        self.assertTrue(reader.test(xml))
        values = reader.read(xml)
        expected = {
            "identifier": "MER_FRS_1PNPDE20060816_090929_000001972050_00222_23322_0058_uint16_reduced_compressed",
            "begin_time": datetime(2006, 8, 16, 9, 9, 29, tzinfo=utc),
            "end_time": datetime(2006, 8, 16, 9, 12, 46, tzinfo=utc),
            "footprint": MultiPolygon(
                Polygon.from_bbox((10, 20, 30, 40)),
                Polygon.from_bbox((50, 60, 70, 80))
            ),
            'format': 'eogml',
            'metadata': {
              'acquisition_station': 'PDHS-E',
              'acquisition_sub_type': None,
              'antenna_look_direction': None,
              'availability_time': datetime(2006, 8, 16, 11, 3, 8, tzinfo=utc),
              'cloud_cover': None,
              'completion_time_from_ascending_node': None,
              'doppler_frequency': None,
              'highest_location': None,
              'illumination_azimuth_angle': None,
              'illumination_elevation_angle': None,
              'illumination_zenith_angle': None,
              'incidence_angle_variation': None,
              'lowest_location': None,
              'maximum_incidence_angle': None,
              'minimum_incidence_angle': None,
              'across_track_incidence_angle': None,
              'along_track_incidence_angle': None,
              'modification_date': None,
              'polarisation_mode': None,
              'polarization_channels': None,
              'snow_cover': None,
              'start_time_from_ascending_node': None}
        }
        self.maxDiff = None
        # self.assertEqual(expected, values)
        values_time_extent = values["begin_time"], values["end_time"]
        expected_time_extent = expected["begin_time"], expected["end_time"]
        self.assertEqual(expected_time_extent, values_time_extent)
        self.assertEqual(expected["identifier"], values["identifier"])
        self.assertEqual(expected["metadata"], values["metadata"])
        self.assertGeometryEqual(expected["footprint"], values["footprint"])

    def test_dimap_reader(self):
        xml = """

        """
        # TODO: find example DIMAP


class CastingTest(TestCase):
    def test_cast(self):
        coverage_type = create(CoverageType,
            name="RGB"
        )
        rectified_grid = Grid.objects.create(
          name="rectified_grid",
          coordinate_reference_system="EPSG:4326",
          axis_1_name="x",
          axis_1_type=0,
          axis_1_offset="5",
          axis_2_name="y",
          axis_2_type=0,
          axis_2_offset="5"
        )
        create(Coverage,
            identifier="rectified-1",
            footprint=GEOSGeometry("MULTIPOLYGON (((-111.6210939999999994 26.8588260000000005, -113.0273439999999994 -4.0786740000000004, -80.6835939999999994 -9.7036739999999995, -68.0273439999999994 15.6088260000000005, -111.6210939999999994 26.8588260000000005)))"),
            begin_time=parse_datetime("2013-06-11T14:55:23Z"),
            end_time=parse_datetime("2013-06-11T14:55:23Z"),
            grid=rectified_grid,
            axis_1_size=100,
            axis_2_size=100,
            coverage_type=coverage_type
        )

        eo_object = EOObject.objects.get(identifier="rectified-1")
        cast_object = cast_eo_object(eo_object)

        self.assertEqual(type(cast_object), Coverage)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging

from eoxserver.backends.cache import setup_cache_session, shutdown_cache_session


logger = logging.getLogger(__name__)


class BackendsCacheMiddleware(object):
    """ A `Django Request Middleware
    <https://docs.djangoproject.com/en/dev/topics/http/middleware/>`_ to manage
    cache setup and teardown when a request is beeing processed.
    """

    def process_request(self, request):
        setup_cache_session()

    def process_response(self, request, response):
        shutdown_cache_session()
        return response

    def process_template_response(self, request, response):
        shutdown_cache_session()
        return response

    def process_exception(self, request, exception):
        shutdown_cache_session()
        return None



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from collections import namedtuple

from django.http import Http404
from django.db.models import Q

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import kvp
from eoxserver.core.util.xmltools import NameSpaceMap
from eoxserver.resources.coverages import models
from eoxserver.services.opensearch.config import (
    get_opensearch_record_model, OpenSearchConfigReader
)
from eoxserver.services.opensearch.formats import get_formats
from eoxserver.services.opensearch.extensions import get_extensions


class SearchContext(namedtuple("SearchContext", [
        "total_count", "start_index", "page_size", "count",
        "parameters", "namespaces"
        ])):

    @property
    def page_count(self):
        divisor = (self.page_size or self.count)
        if divisor == 0:
            return 1

        return self.total_count // divisor

    @property
    def current_page(self):
        divisor = (self.page_size or self.count)
        if divisor == 0:
            return 1

        return self.start_index // divisor


class OpenSearch11SearchHandler(object):
    def handle(self, request, collection_id=None, format_name=None):
        if request.method == "GET":
            request_parameters = request.GET
        elif request.method == "POST":
            request_parameters = request.POST
        else:
            raise Exception("Invalid request method '%s'." % request.method)

        decoder = OpenSearch11BaseDecoder(request_parameters)

        if collection_id:
            # search for products in that collection and coverages not
            # associated with a product but contained in this collection

            ModelClass = get_opensearch_record_model()

            qs = ModelClass.objects.all()
            if ModelClass == models.EOObject:
                qs = qs.filter(
                    Q(product__collections__identifier=collection_id) |
                    Q(
                        coverage__collections__identifier=collection_id,
                        coverage__parent_product__isnull=True
                    )
                ).select_subclasses()
            else:
                qs = qs.filter(collections__identifier=collection_id)

        else:
            qs = models.Collection.objects.all()

        if decoder.search_terms:
            # TODO: search descriptions, summary etc once available
            qs = qs.filter(identifier__icontains=decoder.search_terms)

        namespaces = NameSpaceMap()
        all_parameters = {}
        for search_extension_class in get_extensions():
            # get all search extension related parameters and translate the name
            # to the actual parameter name
            search_extension = search_extension_class()

            params = dict(
                (parameter["type"], request_parameters[parameter["name"]])
                for parameter in search_extension.get_schema(
                    model_class=qs.model
                )
                if parameter["name"] in request_parameters
            )

            qs = search_extension.filter(qs, params)
            namespaces.add(search_extension.namespace)
            all_parameters[search_extension.namespace.prefix] = params

        if not qs.ordered:
            qs = qs.order_by('begin_time')

        # use [:] here, otherwise the queryset would be evaluated and return
        # lists upon slicing
        total_count = qs[:].count()

        # read the configuration and determine the count parameter
        conf = OpenSearchConfigReader(get_eoxserver_config())
        requested_count = min(
            decoder.count if decoder.count is not None else conf.default_count,
            conf.max_count
        )

        start_index = decoder.start_index

        # if count  is zero, then return an 'empty' queryset
        if requested_count == 0:
            qs = models.EOObject.objects.none()
        else:
            qs = qs[start_index:start_index+requested_count]

        result_count = qs[:].count()

        try:
            result_format = next(
                result_format()
                for result_format in get_formats()
                if result_format.name == format_name
            )
        except StopIteration:
            raise Http404("No such result format '%s'." % format_name)

        search_context = SearchContext(
            total_count, start_index,
            requested_count, result_count,
            all_parameters, namespaces
        )

        return (
            result_format.encode(request, collection_id, qs, search_context),
            result_format.mimetype
        )


def pos_int_zero(raw):
    value = int(raw)
    if value < 0:
        raise ValueError("Value is negative")
    return value


def pos_int(raw):
    value = int(raw)
    if value < 1:
        raise ValueError("Value is negative or zero")
    return value


class OpenSearch11BaseDecoder(kvp.Decoder):
    search_terms = kvp.Parameter("q", num="?")
    start_index = kvp.Parameter("startIndex", pos_int_zero, num="?", default=0)
    count = kvp.Parameter("count", pos_int_zero, num="?", default=None)
    output_encoding = kvp.Parameter("outputEncoding", num="?", default="UTF-8")



#-------------------------------------------------------------------------------
#
#  WPS Literal Data - units of measure
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

class UnitOfMeasure(object):
    """ Base unit of measure class.
    The class defines conversion of input values in the given units
    to a common base unit and conversion of the output values from the common
    base unit to the this unit.

    Constructor parameters:
        name    UOM name
    """
    def __init__(self, name):
        self.name = name

    def apply(self, value):
        """ Convert value from the common base to this unit."""
        raise NotImplementedError

    def strip(self, value):
        """ Convert value of this unit to the common base."""
        raise NotImplementedError


class UnitLinear(UnitOfMeasure):
    """ Simple unit of measure with linear conversion (scale and offset):

            value_uom = (value_base - offset)/scale
            value_base = value_uom*scale + offset

    Constructor parameters:
        name    UOM name
        scale   scale factor
        offset  optional base offset (set to 0.0 by default)

    Examples:
        For temperature conversions between the Fahrenheit scale (this UOM)
        and the Kelvin scale (base UOM) set scale to 5.0/9.0 and offset
        to 459.67*5.0/9.0 .

        For simple distance conversions between kilometres (this UOM)
        and metres (base UOM) set scale factor to 1000.0 and offset to 0.0 .
    """

    def __init__(self, name, scale, offset=0):
        UnitOfMeasure.__init__(self, name)
        self._scale = float(scale)
        self._offset = float(offset)
        if self._scale == 0:
            raise ValueError("Invalid zero UOM scale!")

    def apply(self, value):
        """ Convert value from the common base to this unit."""
        return (value - self._offset)/self._scale

    def strip(self, value):
        """ Convert value of this unit to the common base."""
        return value*self._scale + self._offset



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""\
This module imports and initializes GDAL; i.e enables exceptions and registers
all available drivers.
"""

import os
import contextlib


if os.environ.get('READTHEDOCS', None) != 'True':
    try:
        from osgeo.gdal import *
    except ImportError:
        from gdal import *

    try:
        from collections import OrderedDict as SortedDict
    except ImportError:
        from django.utils.datastructures import SortedDict

    UseExceptions()
    AllRegister()

    GCI_TO_NAME = SortedDict((
        (GCI_Undefined, "Undefined"),
        (GCI_GrayIndex, "GrayIndex"),
        (GCI_PaletteIndex, "PaletteIndex"),
        (GCI_RedBand, "RedBand"),
        (GCI_GreenBand, "GreenBand"),
        (GCI_BlueBand, "BlueBand"),
        (GCI_AlphaBand, "AlphaBand"),
        (GCI_HueBand, "HueBand"),
        (GCI_SaturationBand, "SaturationBand"),
        (GCI_LightnessBand, "LightnessBand"),
        (GCI_CyanBand, "CyanBand"),
        (GCI_MagentaBand, "MagentaBand"),
        (GCI_YellowBand, "YellowBand"),
        (GCI_BlackBand, "BlackBand"),
        (GCI_YCbCr_YBand, "YBand"),
        (GCI_YCbCr_CbBand, "CbBand"),
        (GCI_YCbCr_CrBand, "CrBand"),
    ))

    NAME_TO_GCI = dict((j.lower(), i) for (i, j) in GCI_TO_NAME.items())

    GDT_TO_NAME = SortedDict((
        (GDT_Byte, "Byte"),
        (GDT_UInt16, "UInt16"),
        (GDT_Int16, "Int16"),
        (GDT_UInt32, "UInt32"),
        (GDT_Int32, "Int32"),
        (GDT_Float32, "Float32"),
        (GDT_Float64, "Float64"),
        (GDT_CInt16, "CInt16"),
        (GDT_CInt32, "CInt32"),
        (GDT_CFloat32, "CFloat32"),
        (GDT_CFloat64, "CFloat64"),
    ))

    NAME_TO_GDT = SortedDict((j.lower(), i) for (i, j) in GDT_TO_NAME.items())

    GDT_NUMERIC_LIMITS = {
        GDT_Byte: (0, 255),
        GDT_Int16: (-32768, 32767),
        GDT_UInt16: (0, 65535),
        GDT_CInt16: (complex(-32768, -32768), complex(32767, 32767)),
        GDT_Int32: (-2147483648, 2147483647),
        GDT_UInt32: (0, 4294967295),
        GDT_CInt32: (
            complex(-2147483648, -2147483648), complex(2147483647, 2147483647)
        ),
        GDT_Float32: (-3.40282e+38, 3.40282e+38),
        GDT_CFloat32: (
            complex(-3.40282e+38, -3.40282e+38),
            complex(3.40282e+38, 3.40282e+38)
        ),
        GDT_Float64: (-1.79769e+308, 1.79769e+308),
        GDT_CFloat64: (
            complex(-1.79769e+308, -1.79769e+308),
            complex(1.79769e+308, 1.79769e+308)
        )
    }

    GDT_SIGNIFICANT_FIGURES = {
        GDT_Byte: 3,
        GDT_Int16: 5,
        GDT_UInt16: 5,
        GDT_CInt16: 5,
        GDT_Int32: 10,
        GDT_UInt32: 10,
        GDT_CInt32: 10,
        GDT_Float32: 38,
        GDT_CFloat32: 38,
        GDT_Float64: 308,
        GDT_CFloat64: 308
    }

    GDT_INTEGRAL_TYPES = frozenset(
        (GDT_Byte, GDT_Int16, GDT_UInt16, GDT_Int32, GDT_UInt32)
    )

    GDT_INTEGRAL_COMPLEX_TYPES = frozenset((GDT_CInt16, GDT_CInt32))

    GDT_FLOAT_TYPES = frozenset((GDT_Float32, GDT_Float64))

    GDT_FLOAT_COMPLEX_TYPES = frozenset((GDT_CFloat32, GDT_CFloat64))

    GDT_COMPLEX_TYPES = frozenset(
        (GDT_CInt16, GDT_CInt32, GDT_CFloat32, GDT_CFloat64)
    )


def get_extent(ds):
    """ Gets the extent of the GDAL Dataset in the form (min-x, min-y, max-x, max-y).
    """
    gt = ds.GetGeoTransform()

    x_a = gt[0]
    x_b = gt[0] + gt[1] * ds.RasterXSize
    y_a = gt[3]
    y_b = gt[3] + gt[5] * ds.RasterYSize

    return (min(x_a, x_b), min(y_a, y_b), max(x_a, x_b), max(y_a, y_b))


def set_env(env, fail_on_override=False, return_old=False):
    old_values = {} if return_old else None
    for key, value in env.items():
        if fail_on_override or return_old:
            # old_value = GetConfigOption(str(key))
            old_value = os.environ.get(key)
            if fail_on_override and old_value != value:
                raise Exception(
                    'Would override previous value of %s: %s with %s'
                    % (key, old_value, value)
                )
            elif old_value != value:
                old_values[key] = old_value

        # SetConfigOption(str(key), str(value))
        if value is not None:
            os.environ[key] = value

    return old_values


@contextlib.contextmanager
def config_env(env, fail_on_override=False, reset_old=True):
    old_env = set_env(env, fail_on_override, reset_old)
    yield
    if reset_old:
        set_env(old_env, False, False)


def open_with_env(path, env, shared=True):
    with config_env(env, False):
        return OpenShared(path) if shared else Open(path)



#!/usr/bin/env python 
#-----------------------------------------------------------------------
#
# Description: 
#
#   asynchronous processing master daemon
#
#   This is the master server which keeps track of the aynchronous tasks in the 
#   queue and distributes task to the workers 
# 
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@iguassu.cz>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 Iguassu Software Systems, a.s 
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

# default django settings module 
DJANGO_SETTINGS_DEFAULT = "settings"
DJANGO_DB_DEFAULT = "default"

#-------------------------------------------------------------------------------

import os 
import sys 
import signal 
import logging
import traceback
import os.path 
import time 
import struct
import socket 
from datetime import datetime, timedelta 


from multiprocessing import Lock, Process, Queue , cpu_count 
from multiprocessing.queues import Empty as MPQEmpty
from multiprocessing.queues import Full as MPQFull
from django.utils.encoding import smart_text
try:    import cPickle as pickle 
except: import pickle 
try:
    # Python 2
    xrange
except NameError:
    # Python 3, xrange is now named range
    xrange = range
#-------------------------------------------------------------------------------

QUEUE_EMPTY_QUERY_DELAY=1.5 # time in seconds of next query to empty queue 
QUEUE_PUT_TIMEOUT=1.0 # time out used by internal task queue put operation 
QUEUE_CLEAN_UP_COUNT=300 

#-------------------------------------------------------------------------------
# generate unique server instance ID 

SERVER_ID=0 
while 0 == SERVER_ID : 
    tmp = os.urandom(8)
    SERVER_ID = struct.unpack( 'q' , tmp )[0]
    SERVER_ID_STR = "0x%16.16X"%( struct.unpack( 'Q' , tmp )[0] ) 

#-------------------------------------------------------------------------------

dbLock = Lock()
writeLock = Lock()

def write(msg) : 
    writeLock.acquire()
    sys.stdout.write("[%s] %s"%(SERVER_ID_STR,msg) ) 
    writeLock.release()

def debug( msg ) : write( ("DEBUG: %s\n"%(msg)).encode('UTF-8') ) 
def info( msg ) : write( ("INFO: %s\n"%(msg)).encode('UTF-8') ) 
def warn( msg ) : write( ("WARNINIG: %s\n"%(msg)).encode('UTF-8') ) 
def error( msg ) : write( ("ERROR: %s\n"%(msg)).encode('UTF-8') ) 

#-------------------------------------------------------------------------------
# global worker pool

global GWP 

GWP = None 

#-------------------------------------------------------------------------------

# iterrupt and terminate signal handlers

class ExcTerminate( Exception ) : pass  

SS = { signal.SIGINT:"SIGINT" , signal.SIGTERM:"SIGTERM" } 

def signal_handler_dummy(sig, frm): pass

def signal_handler_sigint(sig, frm): 
    global GWP 
    if GWP is not None : 
        GWP.terminate = True 

def signal_handler_sigterm(sig, frm): 
    global GWP 
    if GWP is not None : 
        GWP.terminate = True 
        GWP.killChild = True # force immediate termination 

#-------------------------------------------------------------------------------

class Importer( object ) : 
    """ smart importer of the handler subroutine 
    
        E.g. function 'd()' in module 'a.b.c' 
            given by a path 'a.b.c.d' is imported
    """

    def __init__( self , path ) : 
        """Initialize class from the given handler 
        subroutine dot-path"""

        self.path = path 
        self.module , _ , self.func = path.rpartition(".")
        self.modules = [] 
        self.handler = None 

    def loadHandler( self ) : 
        """ load the handler subroutine """ 
        
        if not self.handler : 

            # initial list of modules 
            ml0 = set( sys.modules ) 
            
            # new list of modules 
            ml1 = set( sys.modules ) 

            self.handler = getattr( __import__( self.module , fromlist=[self.func] ) , self.func ) 

            # store list of loaded modules 
            self.modules = ml1 - ml0 

        return self.handler 

    def unloadHandler( self ) : 
        """ unload the handler subroutine """ 

        if self.handler : 

            self.handler = None 

            # unload the loaded modules 
            for m in self.modules :  
                del( sys.modules[m] ) 

            # store list of loaded modules 
            self.modules = [] 


def taskDispatch( taskID , threadID ) : 
    """ 
        task dispatcher 

        based on the request class the right request hadler is used 
        to process the asynchronous requets 
    """ 
    # status logger 
    pStatus = TaskStatus( taskID , dbLock ) 

    try: 

        # get task parameters 
        requestType , requestID , requestHandler , inputs = dbLocker( dbLock , startTask , taskID ) 

        info( "[%3.3i] PROCESS: %s %s is running ... " % ( threadID , requestType , requestID ) ) 

        # create importer object 
        imp = Importer( requestHandler ) 

        # try to load the right module and handler 
        imp.loadHandler() 

        # execute handler - proper status logging is duty of the callback 
        imp.handler( pStatus , inputs ) 

        # try to unload the handler 
        imp.unloadHandler() 

        # if no terminating status has been set do it right now 
        dbLocker( dbLock , stopTaskSuccessIfNotFinished , taskID )

        info( "[%3.3i] PROCESS: %s %s is finished ... " % ( threadID , requestType , requestID ) ) 

    except (KeyboardInterrupt,SystemExit): raise 
    except Exception as e : 

        pStatus.setFailure( smart_text((e) ) 

        # finish the task 
        error( "[%3.3i] %s " % ( threadID , smart_text((e) ) ) 


#-------------------------------------------------------------------------------


def worker( queue , id ) :
    """ worker function executed by worker subprocesses """ 

    #def signal_handler(sig, frm): raise KeyboardInterrupt
    #signal.signal( signal.SIGINT,  signal_handler )
    # use the dummy hadlers 
    signal.signal( signal.SIGINT,  signal_handler_dummy )
    #signal.signal( signal.SIGTERM, signal_handler_dummy )

    try : 

        while True : 

            try : 
                item = queue.get() 
            except IOError as e : 
                warn( str(e) ) 
                continue 

            # gracefull termination 
            if ( item is None ) : break 

            # run the task 
            taskDispatch( item , id ) 

    except (KeyboardInterrupt,SystemExit): pass  

    info( "[%3.3i] PROCESS: termination " % id )

def cleanup() : 
    """ cleanup function performing reenquing of zombie tasks """ 

    tasks = dbLocker( dbLock , reenqueueZombieTasks , "Reenqueued by ATPD after timeout." ) 

    for (id,task) in tasks : 
        warn( "[MASTER] Task %i:%s renqueued after timeout!"%(id,task) ) 

    tasks = dbLocker( dbLock , deleteRetiredTasks )  

    for (id,task) in tasks : 
        info( "[MASTER] Task %i:%s deleted after expiration!"%(id,task) ) 


class WorkerPool( object ) : 

    def __init__( self , nthread ) :

        self.queue     = Queue( nthread )
        self.terminate = False 
        self.killChild = False 
        self.proces    = [] 

        # start subprocesses 
        for i in xrange( nthread ) : 
            p = Process( target=worker , args=( self.queue , i ) ) 
            p.start() 
            self.proces.append(p)
    

    def __del__( self ) : 

        # if possible process gracefull termination 
        debug( "[MASTER]: enqueueing terminators ... " ) 
        for p in self.proces : 
            self.queue.put( None ) 
            
        if not self.killChild : 
            debug( "[MASTER]: joining subprocesses ... " ) 
            for p in self.proces : 
                p.join() 
        else : 
            debug( "[MASTER]: terminating subprocesses ... " ) 
            for p in self.proces : 
                p.terminate() 


    def startLoop( self ) :  

        # reenqueue hanging tasks
        # TODO: reenqueuePendingTasks() 

        cnt = 0 
        taskIds = [] 
        self.terminate = False 

        while not self.terminate : 

            try: 

                # get a pending task from the queue 
                taskIds = dbLocker( dbLock , dequeueTask , SERVER_ID ) 

            except QueueEmpty : # no task to be processed 

                # perform DB cleanup 
                cleanup()

                # wait some ammount of time 
                time.sleep( QUEUE_EMPTY_QUERY_DELAY ) 

                # clear counter 
                cnt = 0 

                continue 

            # send task to worker 
            for taskId in list(taskIds) : 
                while not self.terminate : 
                    try : self.queue.put(taskId,True,QUEUE_PUT_TIMEOUT) 
                    except MPQFull : continue 
                    taskIds.remove(taskId) 
                    break 

            # increment counter 
            cnt += 1 

            # perform DB cleanup 
            if ( cnt > QUEUE_CLEAN_UP_COUNT ) : 
                cleanup()
                cnt = 0  

        info( "[MASTER]: termination in progress ... " ) 

        # try to reenequeue processes taken from the DB task queue 
        for item in taskIds : 
            debug( "[MASTER]: reenquing task ID=%i ... " % item ) 
            dbLocker( dbLock , reenqueueTask , item , message = "Reenqued by ATPD." ) 
        try: 
            while True : 
                item = self.queue.get(False)
                debug( "[MASTER]: reenquing task ID=%i ... " % item ) 
                dbLocker( dbLock , reenqueueTask , item , message = "Reenqued by ATPD." ) 
        except MPQEmpty : pass 
#-------------------------------------------------------------------------------

def usage() : 
    """ print usage info """ 

    s = [] 
    s.append( "USAGE: %s [-h][-p <directory>][-s <module>][-d <db name>][-n <integer>] " % ( os.path.basename( sys.argv[0] ) ) ) 
    s.append( "" ) 
    s.append( "PARAMETERS: " ) 
    s.append( "    -h   print this info" ) 
    s.append( "    -p   append an addition Python search path (can be repeated)" ) 
    s.append( "    -n   number of worker instance to be started ( N >= 1 , number of CPUs used by default )" ) 
    s.append( "    -s   django settings module (default '%s')"%DJANGO_SETTINGS_DEFAULT ) 
    #s.append( "    -d   django DB name (default '%s')"%DJANGO_DB_DEFAULT ) 
    s.append( "" ) 
    
    return "\n".join(s)
    

#-------------------------------------------------------------------------------

if __name__ == "__main__" : 

    # django settings module 

    DJANGO_SETTINGS = os.environ.get("DJANGO_SETTINGS_MODULE",DJANGO_SETTINGS_DEFAULT) 
    DJANGO_DB       = DJANGO_DB_DEFAULT

    # try to get number of CPUs

    try : 
        NTHREAD = cpu_count()
    except NotImplementedError : 
        NTHREAD = 1 
        warn( "Failed to get number of CPUs! Setting to 1 asynchronous execution thread." ) 

    info( "Default number of working threads: %i" % NTHREAD ) 

    # parse commandline arguments 

    idx = 1 
    while idx < len(sys.argv) : 
        arg = sys.argv[idx] ; idx +=1 ; 
        if arg == '-p' : 
            sys.path.append( sys.argv[idx] ) 
            info("'%s' ... adding to Python search path." % sys.argv[idx] )
            idx += 1 
        elif arg == '-s' : 
            DJANGO_SETTINGS = sys.argv[idx]
            idx += 1 
        elif arg == '-d' : 
            DJANGO_DB = sys.argv[idx]
            idx += 1 
        elif arg == '-n' : 
            NTHREAD = max( 1 , int(sys.argv[idx]) )
            info("Setting number of working threads to: %i" % NTHREAD ) 
            idx += 1
        elif arg == '-h' : 
            sys.stderr.write(usage()) ; sys.exit(0) 
        else : 
            sys.stderr.write(usage()) 
            error( "Invalid commandline option '%s' !" % arg ) 
            sys.exit(1)  

    #-------------------------------------------------------------------
    # initialize the working enviroment 

    # django settings module 
    os.environ["DJANGO_SETTINGS_MODULE"] = DJANGO_SETTINGS 
    info("'%s' ... is set as the Django settings module " % DJANGO_SETTINGS )
    info("'%s' ... is set as the Django database " % DJANGO_DB )

    #usingDB = DJANGO_DB # ???

    # once the search path is set -> load the required modules
    from eoxserver.core.system import System
    from eoxserver.resources.processes.tracker import TaskStatus, QueueEmpty, \
            dequeueTask, startTask, reenqueueTask, stopTaskSuccessIfNotFinished, \
            reenqueueZombieTasks, deleteRetiredTasks, dbLocker 

    # initialize the system 
    System.init()
    #-------------------------------------------------------------------

    info( "ATPD Asynchronous Task Processing Daemon has just been started!")  
    info( "ATPD: id=%s (%i)" % ( SERVER_ID_STR , SERVER_ID ) ) 
    info( "ATPD: hostname=%s" % socket.getfqdn() ) 
    info( "ATPD: pid=%i " % os.getpid() )

    #-------------------------------------------------------------------
    # start the worker pool 
    
    GWP = WorkerPool( NTHREAD ) 

    # use the GWP terminating hadlers 
    signal.signal( signal.SIGINT,  signal_handler_sigint )
    signal.signal( signal.SIGTERM, signal_handler_sigterm )

    # start the main loop 
    GWP.startLoop()

    # use the dummy hadlers 
    signal.signal( signal.SIGINT,  signal_handler_dummy )
    signal.signal( signal.SIGTERM, signal_handler_dummy )

    del GWP



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module contains definition of the auxiliary 2D bounding box class. """


class Rect(tuple):
    """ Named tuple to describe areas in a 2D array like in images. The tuple
        is always in the form (offset_x, offset_y, size_x, size_y).

        :param offset_x: the x offset of the origin
        :param offset_y: the y offset of the origin
        :param size_x: thy x size of the rect
        :param size_y: thy y size of the rect
        :param upper_x: thy upper x offset of the rect (mutually exclusive with
                        size_x)
        :param upper_y: thy upper y offset of the rect (mutually exclusive with
                        size_y)
    """

    __slots__ = ()

    def __new__(cls, offset_x=0, offset_y=0, size_x=None, size_y=None,
                upper_x=0, upper_y=0):

        # To subclass tuples, it is necessary to overwrite the `__new__`
        # method.

        size_x = size_x if size_x is not None else max(0, upper_x - offset_x)
        size_y = size_y if size_y is not None else max(0, upper_y - offset_y)

        return tuple.__new__(cls, (offset_x, offset_y, size_x, size_y))

    offset_x = property(lambda self: self[0])
    offset_y = property(lambda self: self[1])
    offset = property(lambda self: (self.offset_x, self.offset_y))

    size_x = property(lambda self: self[2])
    size_y = property(lambda self: self[3])
    size = property(lambda self: (self.size_x, self.size_y))

    upper_x = property(lambda self: self.offset_x + self.size_x)
    upper_y = property(lambda self: self.offset_y + self.size_y)
    upper = property(lambda self: (self.upper_x, self.upper_y))

    area = property(lambda self: self.size_x * self.size_y)

    def envelope(self, other):
        """ Returns the envelope of two :class:`Rect`, i.e., a smallest
            rectange contaning the input rectangles.
        """
        return Rect(
            offset_x=min(self.offset_x, other[0]),
            offset_y=min(self.offset_y, other[1]),
            upper_x=max(self.upper_x, other[0] + other[2]),
            upper_y=max(self.upper_y, other[1] + other[3])
        )

    __or__ = envelope

    def intersection(self, other):
        """ Returns the intersection of two :class:`Rect`, i.e.,
            a largest common rectanle contained by the input rectangles.
        """

        return Rect(
            offset_x=max(self.offset_x, other[0]),
            offset_y=max(self.offset_y, other[1]),
            upper_x=min(self.upper_x, other[0] + other[2]),
            upper_y=min(self.upper_y, other[1] + other[3])
        )

    __and__ = intersection

    def intersects(self, other):
        """ Tests whether two :class:`Rect` overlap (True) or not (False).
        """
        return self.intersection(other).area > 0

    def translated(self, tup):
        """ Returns a new :class:`Rect` shifted by the given offset.
        """
        return Rect(
            self.offset_x + tup[0], self.offset_y + tup[1],
            self.size_x, self.size_y
        )

    __add__ = translated

    __sub__ = (lambda self, coordinates: self.translated((-coordinates[0], -coordinates[-1])))

    def __repr__(self):
        return "Rect(offset_x=%s, offset_y=%s, size_x=%s, size_y=%s)" % self



#-------------------------------------------------------------------------------
#
#  WPS input and output parameters' base class
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=too-few-public-methods, too-many-arguments

# NOTE: Currently, the inputs parameters are not allowed to be present
#       more that once (maxOccurs=1) per request. These input parameters
#       are, by default, mandatory (minOccur=1). The inputs can be configured
#       as optional (minOccur=0).
#
#       Although not explicitly mentioned by the WPS 1.0.0 standard,
#       it is a common practice that the outputs do not appear more than
#       once in the response (maxOccurs=1).
#       When the explicit specification of the outputs is omitted in the request
#       all process outputs are contained in the default response.

class BaseParamMetadata(object):
    """ Common metadata base of all parameter classes.

    Constructor parameters:
        identifier   item identifier
        title        item title (human-readable name)
        abstract     item abstract (human-readable description)
    """
    def __init__(self, identifier, title=None, abstract=None):
        self.identifier = identifier
        self.title = title
        self.abstract = abstract


class ParamMetadata(BaseParamMetadata):
    """ Common metadata of the execute request parameters.

    Constructor parameters:
        identifier   item identifier
        title        item title (human-readable name)
        abstract     item abstract (human-readable description)
        uom          item LiteralData UOM
        crs          item BoundingBox CRS
        mime_type    item ComplexData mime-type
        encoding     item ComplexData encoding
        schema       item ComplexData schema
    """

    def __init__(self, identifier, title=None, abstract=None, uom=None,
                 crs=None, mime_type=None, encoding=None, schema=None):
        super(ParamMetadata, self).__init__(
            identifier=identifier, title=title, abstract=abstract
        )
        self.uom = uom
        self.crs = crs
        self.mime_type = mime_type
        self.encoding = encoding
        self.schema = schema


class Parameter(BaseParamMetadata):
    """ Base parameter class used by the process definition.

    Constructor parameters:
        identifier  identifier of the parameter.
        title       optional human-readable name (defaults to identifier).
        abstract    optional human-readable verbose description.
        metadata    optional metadata (title/URL dictionary).
        optional    optional boolean flag indicating whether the input
                    parameter is optional or not.
        resolve_input_references Set this option to False not to resolve
                    input references. By default the references are
                    resolved (downloaded and parsed) transparently.
                    If set to False the references must be handled
                    by the process.
    """

    def __init__(self, identifier=None, title=None, abstract=None,
                 metadata=None, optional=False, resolve_input_references=True):
        super(Parameter, self).__init__(
            identifier=identifier, title=title, abstract=abstract
        )
        self.metadata = metadata or {}
        self.is_optional = optional
        self.resolve_input_references = resolve_input_references



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class ConnectorInterface(object):
    """ Interface for connectors between `mapscript.layerObj` and associated
        data.
    """

    def supports(self, data_items):
        """ Returns `True` if the given `data_items` are supported and
            `False` if not.
        """

    def connect(self, coverage, data_items, layer, options):
        """ Connect a layer (a `mapscript.layerObj`) with the given data
            items and coverage (a list of two-tuples: location and semantic).
        """

    def disconnect(self, coverage, data_items, layer, options):
        """ Performs all necessary cleanup operations.
        """


class LayerFactoryInterface(object):
    """ Interface for factories that create `mapscript.layerObj` objects for
        coverages.
    """

    @property
    def suffixes(self):
        """ The suffixes associated with layers this factory produces. This is
            used for "specialized" layers such as "bands" or "outlines" layers.
            For factories that don't use this feature, it can be left out.
        """

    @property
    def requires_connection(self):
        """ Return whether or layers generated by this factory require to be
            connected via a layer connector.
        """

    def generate(self, eo_object, group_layer, options):
        """ Returns an iterable of `mapscript.layerObj` objects preconfigured
            for the given EO object. This is easily done via the `yield`
            statement.
        """

    def generate_group(self, name):
        """ Returns a 'group layer' to be referenced by all other layers
            generated by this factory.
        """


class StyleApplicatorInterface(object):
    """ Interface for style applicators.
    """

    def apply(self, coverage, data_items, layer):
        """ Apply all relevant styles.
        """



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-20 13:00
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('backends', '0001_initial'),
        ('coverages', '0003_metadata_items_semantic'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProductDataItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=1024)),
                ('format', models.CharField(blank=True, max_length=64, null=True)),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='product_data_items', to='coverages.Product')),
                ('storage', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='eoobject',
            name='identifier',
            field=models.CharField(max_length=256, unique=True),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from lxml.builder import ElementMaker

from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.core.util.timetools import isoformat
from eoxserver.resources.coverages import crss

# namespace declarations
ns_gml = NameSpace("http://www.opengis.net/gml/3.2", "gml")
ns_gmlcov = NameSpace("http://www.opengis.net/gmlcov/1.0", "gmlcov")
ns_om = NameSpace("http://www.opengis.net/om/2.0", "om")
ns_eop = NameSpace("http://www.opengis.net/eop/2.0", "eop")

nsmap = NameSpaceMap(ns_gml, ns_gmlcov, ns_om, ns_eop)

# Element factories
GML = ElementMaker(namespace=ns_gml.uri, nsmap=nsmap)
GMLCOV = ElementMaker(namespace=ns_gmlcov.uri, nsmap=nsmap)
OM = ElementMaker(namespace=ns_om.uri, nsmap=nsmap)
EOP = ElementMaker(namespace=ns_eop.uri, nsmap=nsmap)


class GML32Encoder(object):
    def encode_linear_ring(self, ring, sr):
        frmt = "%.3f %.3f" if sr.projected else "%.8f %.8f"

        swap = crss.getAxesSwapper(sr.srid)
        pos_list = " ".join(frmt % swap(*point) for point in ring)

        return GML("LinearRing",
            GML("posList",
                pos_list
            )
        )

    def encode_polygon(self, polygon, base_id):
        return GML("Polygon",
            GML("exterior",
                self.encode_linear_ring(polygon[0], polygon.srs)
            ),
            *(GML("interior",
                self.encode_linear_ring(interior, polygon.srs)
            ) for interior in polygon[1:]),
            **{ns_gml("id"): "polygon_%s" % base_id}
        )

    def encode_multi_surface(self, geom, base_id):
        if geom.geom_typeid in (6, 7):  # MultiPolygon and GeometryCollection
            polygons = [
                self.encode_polygon(polygon, "%s_%d" % (base_id, i+1))
                for i, polygon in enumerate(geom)
            ]
        elif geom.geom_typeid == 3:     # Polygon
            polygons = [self.encode_polygon(geom, base_id)]
        else:
            polygons = []

        return GML("MultiSurface",
            *[GML("surfaceMember", polygon) for polygon in polygons],
            **{ns_gml("id"): "multisurface_%s" % base_id,
               "srsName": "EPSG:%d" % geom.srid
            }
        )

    def encode_time_period(self, begin_time, end_time, identifier):
        return GML("TimePeriod",
            GML("beginPosition", isoformat(begin_time)),
            GML("endPosition", isoformat(end_time)),
            **{ns_gml("id"): identifier}
        )

    def encode_time_instant(self, time, identifier):
        return GML("TimeInstant",
            GML("timePosition", isoformat(time)),
            **{ns_gml("id"): identifier}
        )


class EOP20Encoder(GML32Encoder):
    def encode_footprint(self, footprint, eo_id):
        return EOP("Footprint",
            EOP("multiExtentOf", self.encode_multi_surface(footprint, eo_id)),
            **{ns_gml("id"): "footprint_%s" % eo_id}
        )

    def encode_metadata_property(self, eo_id, contributing_datasets=None):
        return EOP("metaDataProperty",
            EOP("EarthObservationMetaData",
                EOP("identifier", eo_id),
                EOP("acquisitionType", "NOMINAL"),
                EOP("status", "ARCHIVED"),
                *([EOP("composedOf", contributing_datasets)]
                    if contributing_datasets else []
                )
            )
        )

    def encode_earth_observation(self, identifier, begin_time, end_time,
                                 footprint, contributing_datasets=None,
                                 subset_polygon=None):

        if subset_polygon is not None:
            footprint = footprint.intersection(subset_polygon)

        elements = []
        if begin_time and end_time:
            elements.append(
                OM("phenomenonTime",
                    self.encode_time_period(
                        begin_time, end_time, "phen_time_%s" % identifier
                    )
                )
            )
        if end_time:
            elements.append(
                OM("resultTime",
                    self.encode_time_instant(
                        end_time, "res_time_%s" % identifier
                    )
                )
            )

        elements.extend([
            OM("procedure"),
            OM("observedProperty"),
        ])

        if footprint:
            elements.append(
                OM("featureOfInterest",
                    self.encode_footprint(footprint, identifier)
                )
            )
        elements.extend([
            OM("result"),
            self.encode_metadata_property(identifier, contributing_datasets)
        ])

        return EOP("EarthObservation",
            *elements,
            **{ns_gml("id"): "eop_%s" % identifier}
        )



#-------------------------------------------------------------------------------
#
#  CRS data type.
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.resources.coverages.crss import (
    asURL, fromURL, fromURN, fromShortCode, validateEPSGCode, parseEPSGCode,
)
from .data_types import BaseType
from django.utils.encoding import smart_text


class CRSType(BaseType):
    """ CRS data-type.
        CRS are represented by the EPSG codes + 0 meaning the ImageCRC.
    """
    name = "anyURI"
    dtype = int
    zero = None
    comparable = False

    @classmethod
    def parse(cls, raw_value):
        """ Cast or parse input to its proper representation."""
        if isinstance(raw_value, cls.dtype):
            if raw_value == 0 or validateEPSGCode(raw_value):
                return raw_value
        else:
            if raw_value == "ImageCRS":
                return 0
            else:
                value = parseEPSGCode(
                    raw_value, (fromURL, fromURN, fromShortCode)
                )
                if value is not None:
                    return value
        raise ValueError("Invalid CRS %r!" % raw_value)

    @classmethod
    def encode(cls, value):
        """ Encode value to a Unicode string."""
        if value == 0:
            return u'ImageCRS'
        elif validateEPSGCode(value):
            return smart_text(asURL(value))
        raise ValueError("Invalid CRS %r!" % value)

    @classmethod
    def get_diff_dtype(cls):  # string has no difference
        return None



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os
import sys
from optparse import make_option

import django
try:
    # Django versions >= 1.9
    from django.utils.module_loading import import_module
except ImportError:
    # Django versions < 1.9
    from django.utils.importlib import import_module
from django.core.management import BaseCommand
from django.core.management.base import CommandError
from django.utils import termcolors

import eoxserver


class CommandNotFound(Exception):
    def __init__(self, cmdname):
        self.cmdname = cmdname


class EOxServerAdminCommand(BaseCommand):
    def execute(self, *args, **kwargs):
        return self.handle(*args, **kwargs)


def get_commands():
    import eoxserver.core.commands
    command_dir = os.path.dirname(eoxserver.core.commands.__file__)
    command_names = [
        f[:-3] for f in os.listdir(command_dir)
        if not f.startswith('_') and f.endswith('.py')
    ]

    commands = {}
    for name in command_names:
        try:
            module = import_module("eoxserver.core.commands.%s" % name)
            commands[name] = module.Command()

        except ImportError:
            raise

    return commands


def print_possible_commands(commands, stream=sys.stdout):
    stream.write(
        "Type 'eoxserver-admin.py help <subcommand>' "
        "for help on a specific subcommand.\n\n"
    )
    stream.write(
        "Possible commands are:\n" +
        "\t%s" % ("\n\t".join(commands.keys())) +
        "\n"
    )


def execute_from_commandline():
    try:
        subcommand = sys.argv[1]
    except IndexError:
        subcommand = 'help'

    commands = get_commands()

    if subcommand in ('help', '--help', '-h'):
        try:
            cmd = commands[sys.argv[2]]
            cmd.print_help(sys.argv[0], sys.argv[2])
        except IndexError:
            # TODO print general help
            print ("Usage: %s <command-name> [args]" % sys.argv[0])
            print_possible_commands(commands)
        except KeyError:
            print ("Command '%s' not found.\n" % sys.argv[2])
            print_possible_commands(commands)

    elif subcommand == '--version':
        print (eoxserver.get_version())

    else:
        try:
            command = commands[subcommand]
        except KeyError:
            print ("Command '%s' not found.\n" % sys.argv[1])
            print_possible_commands(commands)
            sys.exit(1)

        try:
            command.run_from_argv(sys.argv)
        except CommandError as e:
            print (termcolors.colorize("Error: %s" % e, fg="red"))
            sys.exit(1)



#!/usr/bin/env python
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
Create a new EOxServer instance. This instance will create a root directory
with the instance name in the given (optional) directory.
"""

from optparse import make_option

from django.core.management.base import CommandError

from eoxserver.core.management import EOxServerAdminCommand
from eoxserver.core.instance import create_instance


class Command(EOxServerAdminCommand):
    def add_arguments(self, parser):
        parser.add_argument('instance_id',
            metavar='INSTANCE_ID', nargs=1,
            help='The instance idnetifier.'
        )
        parser.add_argument('target',
            metavar='DEST_DIRECTORY', nargs='?',
            help='Optional. The destination directory.'
        )
        parser.add_argument('--init_spatialite', '--init-spatialite',
            action='store_true', help='Flag to initialize the sqlite database.'
        )

    args = "INSTANCE_ID [Optional destination directory] [--init-spatialite]"
    help = ("Creates a new EOxServer instance with all necessary files and "
            "folder structure. Optionally, a SQLite database is initiated")

    def handle(self, instance_id, target=None, *args, **options):
        if not instance_id:
            raise CommandError("Instance ID not given.")

        init_spatialite = options["init_spatialite"]
        verbosity = options["verbosity"]
        traceback = options.get("traceback")

        create_instance(
            instance_id[0], target, init_spatialite, verbosity, traceback
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from itertools import chain

from eoxserver.core import Component, implements, ExtensionPoint
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.util.timetools import isoformat
from eoxserver.contrib.mapserver import create_request, Map, Layer, Class, Style
from eoxserver.resources.coverages import crss, models
from eoxserver.resources.coverages.formats import getFormatRegistry
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.wms.interfaces import (
    WMSCapabilitiesRendererInterface
)
from eoxserver.services.mapserver.interfaces import LayerFactoryInterface
from eoxserver.services.result import result_set_from_raw_data, get_content_type
from eoxserver.services.urls import get_http_service_url


class MapServerWMSCapabilitiesRenderer(Component):
    """ WMS Capabilities renderer implementation using MapServer.
    """
    implements(WMSCapabilitiesRendererInterface)

    layer_factories = ExtensionPoint(LayerFactoryInterface)

    @property
    def suffixes(self):
        return list(
            chain(*[factory.suffixes for factory in self.layer_factories])
        )

    def render(self, collections, coverages, request_values, request):
        conf = CapabilitiesConfigReader(get_eoxserver_config())

        suffixes = self.suffixes

        http_service_url = get_http_service_url(request)

        map_ = Map()
        map_.setMetaData({
            "enable_request": "*",
            "onlineresource": http_service_url,
            "service_onlineresource": conf.onlineresource,
            "updateSequence": conf.update_sequence,
            "name": conf.name,
            "title": conf.title,
            "abstract": conf.abstract,
            "accessconstraints": conf.access_constraints,
            "addresstype": "postal",
            "address": conf.delivery_point,
            "stateorprovince": conf.administrative_area,
            "city": conf.city,
            "postcode": conf.postal_code,
            "country": conf.country,
            "contactelectronicmailaddress": conf.electronic_mail_address,
            "contactfacsimiletelephone": conf.phone_facsimile,
            "contactvoicetelephone": conf.phone_voice,
            "contactperson": conf.individual_name,
            "contactorganization": conf.provider_name,
            "contactposition": conf.position_name,
            "fees": conf.fees,
            "keywordlist": ",".join(conf.keywords),
            "srs": " ".join(crss.getSupportedCRS_WCS(format_function=crss.asShortCode)),
        }, namespace="ows")
        map_.setProjection("EPSG:4326")
        map_.setMetaData({
            "getmap_formatlist": ",".join([f.mimeType for f in self.get_wms_formats()]),
            "getfeatureinfo_formatlist": "text/html,application/vnd.ogc.gml,text/plain",
        }, namespace="wms")

        map_extent = None

        for collection in collections:
            group_name = None
            
            # calculate extent and timextent for every collection
            extent = collection.extent_wgs84
            # save overall map extent
            map_extent = self.join_extents(map_extent, extent)

            eo_objects = collection.eo_objects.filter(
                begin_time__isnull=False, end_time__isnull=False
            )
            timeextent = ",".join(
                map(
                    lambda o: (
                        "/".join(
                            map(isoformat, o.time_extent)
                        ) + "/PT1S"
                    ), eo_objects
                )
            )

            if len(suffixes) > 1:
                # create group layer, if there is more than one suffix for this 
                # collection
                group_name = collection.identifier + "_group"
                group_layer = Layer(group_name)
                group_layer.setMetaData({
                    "title": group_name,
                    "abstract": group_name,
                    "extent": " ".join(map(str, extent)),
                }, namespace="wms")

                minx, miny, maxx, maxy = extent
                group_layer.setExtent(minx, miny, maxx, maxy)

                # add default style
                default_class = Class("default")
                default_style= Style("default")
                default_class.insertStyle(default_style)
                group_layer.insertClass(default_class)

                map_.insertLayer(group_layer)

            for suffix in suffixes:
                layer_name = collection.identifier + (suffix or "")
                layer = Layer(layer_name)
                if group_name:
                    layer.setMetaData({
                        "layer_group": "/" + group_name
                    }, namespace="wms")

                layer.setMetaData({
                    "title": layer_name,
                    "abstract": layer_name,
                    "extent": " ".join(map(str, extent)),
                    "timeextent": timeextent,
                }, namespace="wms")
                map_.insertLayer(layer)

        for coverage in coverages:
            extent = coverage.extent_wgs84
            minx, miny, maxx, maxy = extent
            # save overall map extent
            map_extent = self.join_extents(map_extent, extent)

            layer_name = coverage.identifier
            layer = Layer(layer_name)
            layer.setMetaData({
                "title": layer_name,
                "abstract": layer_name,
                "extent": " ".join(map(str, extent)),
            }, namespace="wms")
            minx, miny, maxx, maxy = extent
            layer.setExtent(minx, miny, maxx, maxy)

            map_.insertLayer(layer)

        # set the map_extent to a reasonable default value
        # in case there is no coverage or collection
        if map_extent is None : 
            map_extent = ( 0.0, 0.0, 1.0, 1.0 )

        map_minx, map_miny, map_maxx, map_maxy = map_extent
        map_.setExtent(map_minx, map_miny, map_maxx, map_maxy)

        request = create_request(request_values)
        raw_result = map_.dispatch(request)
        result = result_set_from_raw_data(raw_result)
        return result, get_content_type(result)

    def get_wms_formats(self):
        return getFormatRegistry().getSupportedFormatsWMS()

    def join_extents(self, e1=None, e2=None):
        if e1 and e2:
            e1_minx, e1_miny, e1_maxx, e1_maxy = e1
            e2_minx, e2_miny, e2_maxx, e2_maxy = e2
            return (
                min(e1_minx, e2_minx), 
                min(e1_miny, e2_miny), 
                max(e1_maxx, e2_maxx), 
                max(e1_maxy, e2_maxy)
            )
        elif e1:
            return e1
        elif e2:
            return e2
        else:
            return None



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from django.contrib.gis.geos import Polygon

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.resources.coverages.metadata.utils.landsat8_l1 import (
    is_landsat8_l1_metadata_file, parse_landsat8_l1_metadata_file
)


class Landsat8L1ProductMetadataReader(object):
    def test_path(self, path):
        return is_landsat8_l1_metadata_file(path)

    def read_path(self, path):
        md = parse_landsat8_l1_metadata_file(path)

        p = md['PRODUCT_METADATA']
        ul = float(p['CORNER_UL_LON_PRODUCT']), float(p['CORNER_UL_LAT_PRODUCT'])
        ur = float(p['CORNER_UR_LON_PRODUCT']), float(p['CORNER_UR_LAT_PRODUCT'])
        ll = float(p['CORNER_LL_LON_PRODUCT']), float(p['CORNER_LL_LAT_PRODUCT'])
        lr = float(p['CORNER_LR_LON_PRODUCT']), float(p['CORNER_LR_LAT_PRODUCT'])

        values = {}
        values['identifier'] = md['METADATA_FILE_INFO']['LANDSAT_SCENE_ID']
        values['footprint'] = Polygon([ul, ur, lr, ll, ul])
        time = parse_iso8601('%sT%s' % (
            p['DATE_ACQUIRED'], p['SCENE_CENTER_TIME']
        ))
        values['begin_time'] = values['end_time'] = time
        values['cloud_cover'] = float(md['IMAGE_ATTRIBUTES']['CLOUD_COVER'])
        values['track'] = p['WRS_PATH']
        values['frame'] = p['WRS_ROW']

        values['processing_date'] = parse_iso8601(
            md['METADATA_FILE_INFO']['FILE_DATE']
        )

        # TODO: maybe convert additional fields from Metadata file

        return values

        # from pprint import pprint; pprint(values)

        # values['parent_identifier']
        # values['production_status']
        # values['acquisition_type']
        # values['orbit_number'] = ds.sensing_orbit_number
        # values['orbit_direction'] = ds.sensing_orbit_dir
        # values['track']
        # values['frame']
        # values['swath_identifier'] = metadata.find('.//P
        # values['product_version'] = metadata.findtext('.
        # values['product_quality_status']
        # values['product_quality_degradation_tag']
        # values['processor_name']
        # values['processing_center']
        # values['creation_date']
        # values['modification_date']
        # values['processing_date'] = ds.generation_time
        # values['sensor_mode']
        # values['archiving_center'] = granule_metadata.fi
        # values['processing_mode']
        # values['availability_time'] = ds.generation_time
        # values['acquisition_station']
        # values['acquisition_sub_type']
        # values['start_time_from_ascending_node']
        # values['completion_time_from_ascending_node']
        # values['illumination_azimuth_angle'] = metadata.
        # values['illumination_zenith_angle'] = metadata.f
        # values['illumination_elevation_angle']
        # values['polarisation_mode']
        # values['polarization_channels']
        # values['antenna_look_direction']
        # values['minimum_incidence_angle']
        # values['maximum_incidence_angle']
        # values['doppler_frequency']
        # values['incidence_angle_variation']
        # values['cloud_cover'] = metadata.findtext(".//Cl
        # values['snow_cover']
        # values['lowest_location']
        # values['highest_location']



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from __future__ import division

from itertools import chain
from lxml import etree

from django.contrib.gis.geos import Polygon
from django.utils.timezone import now

from eoxserver.contrib import gdal, vsi
from eoxserver.core.util.timetools import isoformat
from eoxserver.backends.access import retrieve
from eoxserver.contrib.osr import SpatialReference
from eoxserver.render.coverage import objects
from eoxserver.resources.coverages.formats import getFormatRegistry
from eoxserver.resources.coverages import crss
from eoxserver.services.gml.v32.encoders import GML32Encoder, EOP20Encoder
from eoxserver.services.ows.component import ServiceComponent, env
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.common.v20.encoders import OWS20Encoder
from eoxserver.services.ows.wcs.v20.util import (
    nsmap, ns_xlink, ns_gml, ns_wcs, ns_eowcs,
    OWS, GML, GMLCOV, WCS, CRS, EOWCS, SWE, INT, SUPPORTED_INTERPOLATIONS
)

PROFILES = [
    "spec/WCS_application-profile_earth-observation/1.0/conf/eowcs",
    "spec/WCS_application-profile_earth-observation/1.0/conf/eowcs_get-kvp",
    "spec/WCS_service-extension_crs/1.0/conf/crs",
    "spec/WCS/2.0/conf/core",
    "spec/WCS_protocol-binding_get-kvp/1.0/conf/get-kvp",
    "spec/WCS_protocol-binding_post-xml/1.0/conf/post-xml",
    "spec/GMLCOV/1.0/conf/gml-coverage",
    "spec/GMLCOV/1.0/conf/multipart",
    "spec/GMLCOV/1.0/conf/special-format",
    "spec/GMLCOV_geotiff-coverages/1.0/conf/geotiff-coverage",
    "spec/WCS_geotiff-coverages/1.0/conf/geotiff-coverage",
    "spec/WCS_service-model_crs-predefined/1.0/conf/crs-predefined",
    "spec/WCS_service-extension_interpolation/1.0/conf/interpolation",
    "spec/WCS_service-extension_range-subsetting/1.0/conf/record-subsetting",
    "spec/WCS_service-extension_scaling/1.0/conf/scaling",
]


class WCS20BaseXMLEncoder(object):
    def get_coverage_subtype(self, coverage):
        if isinstance(coverage, objects.Mosaic):
            return "RectifiedStitchedMosaic"
        elif coverage.grid and coverage.grid.is_referenceable:
            return "ReferenceableDataset"
        elif not coverage.footprint or not coverage.begin_time or \
                not coverage.end_time:
            return "RectifiedGridCoverage"

        return "RectifiedDataset"


class WCS20CapabilitiesXMLEncoder(WCS20BaseXMLEncoder, OWS20Encoder):
    def get_conf(self):
        from eoxserver.core.config import get_eoxserver_config
        return CapabilitiesConfigReader(get_eoxserver_config())

    def encode_service_metadata(self):
        service_metadata = WCS("ServiceMetadata")

        # get the list of enabled formats from the format registry
        formats = filter(
            lambda f: f, getFormatRegistry().getSupportedFormatsWCS()
        )
        service_metadata.extend(
            map(lambda f: WCS("formatSupported", f.mimeType), formats)
        )

        # get a list of supported CRSs from the CRS registry
        supported_crss = crss.getSupportedCRS_WCS(
            format_function=crss.asURL
        )
        extension = WCS("Extension")
        service_metadata.append(extension)
        crs_metadata = CRS("CrsMetadata")
        extension.append(crs_metadata)
        crs_metadata.extend(
            map(lambda c: CRS("crsSupported", c), supported_crss)
        )

        base_url = "http://www.opengis.net/def/interpolation/OGC/1/"

        extension.append(
            INT("InterpolationMetadata", *[
                INT("InterpolationSupported",
                    base_url + supported_interpolation
                ) for supported_interpolation in SUPPORTED_INTERPOLATIONS
            ])
        )
        return service_metadata

    def encode_contents(self, coverages, dataset_series_set):
        contents = []

        if coverages is not None:
            contents.extend([
                WCS("CoverageSummary",
                    WCS("CoverageId", coverage.identifier),
                    WCS("CoverageSubtype",
                        self.get_coverage_subtype(coverage)
                    )
                ) for coverage in coverages
            ])

        if dataset_series_set is not None:
            dataset_series_elements = []
            for dataset_series in dataset_series_set:
                footprint = dataset_series.footprint
                dataset_series_summary = EOWCS("DatasetSeriesSummary")

                # NOTE: non-standard, ows:WGS84BoundingBox is actually mandatory,
                # but not available for e.g: empty collections
                if footprint:
                    minx, miny, maxx, maxy = footprint.extent
                    dataset_series_summary.append(
                        OWS("WGS84BoundingBox",
                            OWS("LowerCorner", "%f %f" % (miny, minx)),
                            OWS("UpperCorner", "%f %f" % (maxy, maxx)),
                        )
                    )

                dataset_series_summary.append(
                    EOWCS("DatasetSeriesId", dataset_series.identifier)
                )

                # NOTE: non-standard, gml:TimePosition is actually mandatory,
                # but not available for e.g: empty collections
                if dataset_series.begin_time and dataset_series.end_time:
                    dataset_series_summary.append(
                        GML("TimePeriod",
                            GML(
                                "beginPosition",
                                isoformat(dataset_series.begin_time)
                            ),
                            GML(
                                "endPosition",
                                isoformat(dataset_series.end_time)
                            ),
                            **{
                                ns_gml("id"): dataset_series.identifier +
                                "_timeperiod"
                            }
                        )
                    )

                dataset_series_elements.append(dataset_series_summary)

            if dataset_series_elements:
                contents.append(WCS("Extension", *dataset_series_elements))

        return WCS("Contents", *contents)

    def encode_capabilities(self, sections, conf, coverages=None,
                            dataset_series=None, request=None,):

        conf = self.get_conf()

        all_sections = "all" in sections
        caps = []
        if all_sections or "serviceidentification" in sections:
            caps.append(self.encode_service_identification(
                "WCS", conf, PROFILES
            ))

        if all_sections or "serviceprovider" in sections:
            caps.append(self.encode_service_provider(conf))

        if all_sections or "operationsmetadata" in sections:
            caps.append(self.encode_operations_metadata(
                request, "WCS", ("2.0.0", "2.0.1")
            ))

        if all_sections or "servicemetadata" in sections:
            caps.append(self.encode_service_metadata())

        inc_contents = all_sections or "contents" in sections
        inc_coverage_summary = inc_contents or "coveragesummary" in sections
        inc_dataset_series_summary = (
            inc_contents or "datasetseriessummary" in sections
        )

        if inc_contents or inc_coverage_summary or inc_dataset_series_summary:
            caps.append(
                self.encode_contents(
                    coverages if inc_coverage_summary else None,
                    dataset_series if inc_dataset_series_summary else None,
                )
            )

        return WCS(
            "Capabilities", *caps, version="2.0.1",
            updateSequence=conf.update_sequence
        )

    def get_schema_locations(self):
        return nsmap.schema_locations


class GMLCOV10Encoder(WCS20BaseXMLEncoder, GML32Encoder):
    def __init__(self, *args, **kwargs):
        self._cache = {}

    def get_gml_id(self, identifier):
        if identifier[0].isdigit():
            return "gmlid_%s" % identifier
        return identifier

    def encode_grid_envelope(self, sizes):
        return GML("GridEnvelope",
            GML("low", " ".join("0" for size in sizes)),
            GML("high", " ".join(("%d" % (size - 1) for size in sizes)))
        )

    def encode_rectified_grid(self, grid, coverage, name):
        axis_names = [axis.name for axis in grid]
        offsets = [axis.offset for axis in grid]
        origin = coverage.origin

        sr = SpatialReference(grid.coordinate_reference_system)
        url = sr.url

        frmt = "%.3f" if sr.IsProjected() else "%.8f"
        offset_vectors = [
            GML("offsetVector",
                " ".join(
                    [frmt % 0.0] * i + [frmt % offset]
                    + [frmt % 0.0] * (len(offsets) - i - 1)
                ),
                srsName=url
            )
            for i, offset in enumerate(offsets)
        ]

        if crss.hasSwappedAxes(sr.srid):
            axis_names[0:2] = [axis_names[1], axis_names[0]]
            # offset_vectors[0:2] = [offset_vectors[1], offset_vectors[0]]
            for offset_vector in offset_vectors[0:2]:
                parts = offset_vector.text.split(" ")
                parts[0:2] = reversed(parts[0:2])
                offset_vector.text = " ".join(parts)

            origin[0:2] = [origin[1], origin[0]]

        origin_str = " ".join(
            ["%.3f" if sr.IsProjected() else "%.8f"] * len(origin)
        ) % tuple(origin)

        return GML("RectifiedGrid",
            GML("limits",
                self.encode_grid_envelope(coverage.size)
            ),
            GML("axisLabels", " ".join(axis_names)),
            GML("origin",
                GML("Point",
                    GML("pos", origin_str),
                    **{
                        ns_gml("id"): self.get_gml_id("%s_origin" % name),
                        "srsName": url
                    }
                )
            ),
            *offset_vectors,
            **{
                ns_gml("id"): self.get_gml_id(name),
                "dimension": "2"
            }
        )

    def encode_referenceable_grid(self, size, sr, grid_name):
        size_x, size_y = size
        swap = crss.getAxesSwapper(sr.srid)
        labels = ("x", "y") if sr.IsProjected() else ("long", "lat")
        axis_labels = " ".join(swap(*labels))

        return GML("ReferenceableGrid",
            GML("limits",
                self.encode_grid_envelope([size_x - 1, size_y - 1])
            ),
            GML("axisLabels", axis_labels),
            **{
                ns_gml("id"): self.get_gml_id(grid_name),
                "dimension": "2"
            }
        )

    def encode_domain_set(self, coverage, srid=None, size=None, extent=None,
                          rectified=True):
        grid_name = "%s_grid" % coverage.identifier
        grid = coverage.grid
        sr = SpatialReference(grid.coordinate_reference_system)

        if grid and not grid.is_referenceable:
            return GML("domainSet",
                self.encode_rectified_grid(
                    grid, coverage, grid_name
                )
            )
        else:
            return GML("domainSet",
                self.encode_referenceable_grid(
                    size or coverage.size,
                    sr,
                    grid_name
                )
            )

    def encode_bounded_by(self, coverage, grid=None, subset_extent=None):
        # if grid is None:
        footprint = coverage.footprint

        if grid and not grid.is_referenceable:
            sr = SpatialReference(grid.coordinate_reference_system)
            labels = grid.names
            axis_units = " ".join(
                ["m" if sr.IsProjected() else "deg"] * len(labels)
            )
            extent = list(subset_extent) if subset_extent else list(coverage.extent)

            lc = extent[:len(extent) // 2]
            uc = extent[len(extent) // 2:]

            if crss.hasSwappedAxes(sr.srid):
                labels[0:2] = labels[1], labels[0]
                lc[0:2] = lc[1], lc[0]
                uc[0:2] = uc[1], uc[0]

            frmt = " ".join(
                ["%.3f" if sr.IsProjected() else "%.8f"] * len(labels)
            )

            lower_corner = frmt % tuple(lc)
            upper_corner = frmt % tuple(uc)
            axis_labels = " ".join(labels)
            srs_name = sr.url

        elif footprint:
            minx, miny, maxx, maxy = subset_extent or footprint.extent
            sr = SpatialReference(4326)
            swap = crss.getAxesSwapper(sr.srid)
            labels = ("x", "y") if sr.IsProjected() else ("long", "lat")
            axis_labels = " ".join(swap(*labels))
            axis_units = "m m" if sr.IsProjected() else "deg deg"
            frmt = "%.3f %.3f" if sr.IsProjected() else "%.8f %.8f"

            # Make sure values are outside of actual extent
            if sr.IsProjected():
                minx -= 0.0005
                miny -= 0.0005
                maxx += 0.0005
                maxy += 0.0005
            else:
                minx -= 0.000000005
                miny -= 0.000000005
                maxx += 0.000000005
                maxy += 0.000000005

            lower_corner = frmt % swap(minx, miny)
            upper_corner = frmt % swap(maxx, maxy)
            srs_name = sr.url

        else:
            lower_corner = ""
            upper_corner = ""
            srs_name = ""
            axis_labels = ""
            axis_units = ""

        return GML("boundedBy",
            GML("Envelope",
                GML("lowerCorner", lower_corner),
                GML("upperCorner", upper_corner),
                srsName=srs_name, axisLabels=axis_labels, uomLabels=axis_units,
                srsDimension="2"
            )
        )

    def encode_nil_values(self, nil_values):
        return SWE("nilValues",
            SWE("NilValues",
                *[
                    SWE("nilValue", nil_value[0], reason=nil_value[1])
                    for nil_value in nil_values
                ]
            )
        )

    def encode_field(self, field):
        return SWE("field",
            SWE("Quantity",
                SWE("description", field.description),
                self.encode_nil_values(field.nil_values),
                SWE("uom", code=field.unit_of_measure),
                SWE("constraint",
                    SWE("AllowedValues",
                        *[
                            SWE("interval", "%g %g" % value_range)
                            for value_range in field.allowed_values
                        ] + [
                            SWE("significantFigures", str(
                                field.significant_figures
                            ))
                        ] if field.significant_figures else []
                    )
                ),
                # TODO: lookup correct definition according to data type:
                # http://www.opengis.net/def/dataType/OGC/0/
                definition=field.definition
            ),
            name=field.identifier
        )

    def encode_range_type(self, range_type):
        return GMLCOV("rangeType",
            SWE("DataRecord",
                *[self.encode_field(band) for band in range_type]
            )
        )


class WCS20CoverageDescriptionXMLEncoder(GMLCOV10Encoder):
    def encode_coverage_description(self, coverage):
        grid = coverage.grid
        return WCS("CoverageDescription",
            self.encode_bounded_by(coverage, grid),
            WCS("CoverageId", coverage.identifier),
            self.encode_domain_set(coverage, rectified=(grid is not None)),
            self.encode_range_type(coverage.range_type),
            WCS("ServiceParameters",
                WCS("CoverageSubtype", self.get_coverage_subtype(coverage))
            ),
            **{ns_gml("id"): self.get_gml_id(coverage.identifier)}
        )

    def encode_coverage_descriptions(self, coverages):
        return WCS("CoverageDescriptions", *[
            self.encode_coverage_description(coverage)
            for coverage in coverages
        ])

    def get_schema_locations(self):
        return {ns_wcs.uri: ns_wcs.schema_location}


class WCS20EOXMLEncoder(WCS20CoverageDescriptionXMLEncoder, EOP20Encoder,
                        OWS20Encoder):
    def encode_eo_metadata(self, coverage, request=None, subset_polygon=None):
        metadata_items = [
            metadata_location
            for metadata_location in getattr(coverage, 'metadata_locations', [])
            if metadata_location.format == "eogml"
        ]
        if len(metadata_items) >= 1:
            with vsi.open(metadata_items[0].path) as f:
                earth_observation = etree.parse(f).getroot()

            if subset_polygon:
                try:
                    feature = earth_observation.xpath(
                        "om:featureOfInterest", namespaces=nsmap
                    )[0]
                    feature[0] = self.encode_footprint(
                        coverage.footprint.intersection(subset_polygon),
                        coverage.identifier
                    )
                except IndexError:
                    pass  # no featureOfInterest

        else:
            earth_observation = self.encode_earth_observation(
                coverage.identifier, coverage.begin_time, coverage.end_time,
                coverage.footprint, subset_polygon=subset_polygon
            )

        if not request:
            lineage = None

        elif request.method == "GET":
            lineage = EOWCS("lineage",
                EOWCS("referenceGetCoverage",
                    self.encode_reference("Reference",
                        request.build_absolute_uri().replace("&", "&amp;"),
                        False
                    )
                ), GML("timePosition", isoformat(now()))
            )
        elif request.method == "POST":  # TODO: better way to do this
            href = request.build_absolute_uri().replace("&", "&amp;")
            lineage = EOWCS("lineage",
                EOWCS("referenceGetCoverage",
                    OWS("ServiceReference",
                        OWS("RequestMessage",
                            etree.parse(request).getroot()
                        ), **{ns_xlink("href"): href}
                    )
                ), GML("timePosition", isoformat(now()))
            )

        return GMLCOV("metadata",
            GMLCOV("Extension",
                EOWCS("EOMetadata",
                    earth_observation,
                    *[lineage] if lineage is not None else []
                )
            )
        )

    def encode_coverage_description(self, coverage, srid=None, size=None,
                                    extent=None, footprint=None):
        source_mime = None
        for arraydata_location in getattr(coverage, 'arraydata_locations', []):
            if arraydata_location.format:
                source_mime = arraydata_location.format
                break

        native_format = None
        if source_mime:
            source_format = getFormatRegistry().getFormatByMIME(source_mime)
            # map the source format to the native one
            native_format = getFormatRegistry().mapSourceToNativeWCS20(
                source_format
            )
        # elif issubclass(coverage.real_type, RectifiedStitchedMosaic):
        #     # use the default format for RectifiedStitchedMosaics
        #     native_format = getFormatRegistry().getDefaultNativeFormat()
        # else:
        #     # TODO: improve if no native format availabe
        #     native_format = None
        sr = SpatialReference(4326)
        if extent:
            poly = Polygon.from_bbox(extent)
            poly.srid = srid
            extent = poly.transform(4326).extent

        else:
            # extent = coverage.extent
            extent = (0, 0, 1, 1)
            # sr = coverage.spatial_reference

        # if issubclass(coverage.real_type, ReferenceableDataset):
        #     rectified = False
        # else:
        #     rectified = True

        rectified = (coverage.grid is not None)

        return WCS("CoverageDescription",
            self.encode_bounded_by(coverage, coverage.grid),
            WCS("CoverageId", coverage.identifier),
            self.encode_eo_metadata(coverage),
            self.encode_domain_set(coverage, srid, size, extent, rectified),
            self.encode_range_type(coverage.range_type),
            WCS("ServiceParameters",
                WCS("CoverageSubtype", self.get_coverage_subtype(coverage)),
                WCS(
                    "nativeFormat",
                    native_format.mimeType if native_format else ""
                )
            ),
            **{ns_gml("id"): self.get_gml_id(coverage.identifier)}
        )

    def encode_range_set(self, reference, mime_type):
        return GML("rangeSet",
            GML("File",
                GML("rangeParameters",
                    **{
                        ns_xlink("arcrole"): "fileReference",
                        ns_xlink("href"): reference,
                        ns_xlink("role"): "http://www.opengis.net/spec/GMLCOV_geotiff-coverages/1.0/conf/geotiff-coverage"
                    }
                ),
                GML("fileReference", reference),
                GML("fileStructure"),
                GML("mimeType", mime_type)
            )
        )

    def calculate_contribution(self, footprint, contributions,
                               subset_polygon=None):
        if subset_polygon:
            footprint = footprint.intersection(subset_polygon)

        for contribution in contributions:
            footprint = footprint.difference(contribution)
        contributions.append(footprint)
        return footprint

    def encode_contributing_datasets(self, mosaic, subset_polygon=None):
        actual_contributions = []
        all_contributions = []
        for coverage in mosaic.coverages:
            contribution = self.calculate_contribution(
                coverage.footprint, all_contributions, subset_polygon
            )
            if not contribution.empty and contribution.num_geom > 0:
                actual_contributions.append((coverage, contribution))

        return EOWCS("datasets", *[
            EOWCS("dataset",
                WCS("CoverageId", coverage.identifier),
                EOWCS("contributingFootprint",
                    self.encode_footprint(
                        contrib, coverage.identifier
                    )
                )
            )
            for coverage, contrib in reversed(actual_contributions)
        ])

    def alter_rectified_dataset(self, coverage, request, tree,
                                subset_polygon=None):
        return EOWCS("RectifiedDataset", *(
            tree.getchildren() + [
                self.encode_eo_metadata(coverage, request, subset_polygon)
            ]
        ), **tree.attrib)

    def alter_rectified_stitched_mosaic(self, coverage, request, tree,
                                        subset_polygon=None):
        return EOWCS("RectifiedStitchedMosaic", *(
            tree.getchildren() + [
                self.encode_eo_metadata(coverage, request, subset_polygon),
                self.encode_contributing_datasets(coverage, subset_polygon)
            ]
        ), **tree.attrib)

    def encode_rectified_dataset(self, coverage, request, reference,
                                 mime_type, subset_polygon=None):
        return EOWCS("RectifiedDataset",
            self.encode_bounded_by(coverage, coverage.grid),
            self.encode_domain_set(coverage, rectified=True),
            self.encode_range_set(reference, mime_type),
            self.encode_range_type(coverage.range_type),
            self.encode_eo_metadata(coverage, request, subset_polygon),
            **{
                ns_gml("id"): self.get_gml_id(coverage.identifier)
            }
        )

    def encode_rectified_stitched_mosaic(self, coverage, request, reference,
                                         mime_type, subset_polygon=None):
        return EOWCS("RectifiedStitchedMosaic",
            self.encode_bounded_by(coverage, coverage.grid),
            self.encode_domain_set(coverage, rectified=True),
            self.encode_range_set(reference, mime_type),
            self.encode_range_type(coverage.range_type),
            self.encode_eo_metadata(coverage, request, subset_polygon),
            self.encode_contributing_datasets(coverage, subset_polygon),
            **{
                ns_gml("id"): self.get_gml_id(coverage.identifier)
            }
        )

    def encode_referenceable_dataset(self, coverage, range_type, reference,
                                     mime_type, subset=None):
        # handle subset
        dst_srid = SpatialReference(coverage.grid.coordinate_reference_system).srid

        if not subset:
            # whole area - no subset
            domain_set = self.encode_domain_set(coverage, rectified=False)
            eo_metadata = self.encode_eo_metadata(coverage)
            extent = coverage.footprint.extent
            sr = SpatialReference(dst_srid)

        else:
            # subset is given
            srid, size, extent, footprint = subset
            srid = srid if srid is not None else 4326

            domain_set = self.encode_domain_set(
                coverage, srid, size, extent, False
            )
            eo_metadata = self.encode_eo_metadata(
                coverage, subset_polygon=footprint
            )

            # get the WGS84 extent
            poly = Polygon.from_bbox(extent)
            poly.srid = srid
            if srid != dst_srid:
                poly.transform(dst_srid)
            extent = poly.extent
            sr = SpatialReference(srid)

        return EOWCS("ReferenceableDataset",
            self.encode_bounded_by(coverage, coverage.grid, extent),
            domain_set,
            self.encode_range_set(reference, mime_type),
            self.encode_range_type(range_type),
            eo_metadata,
            **{
                ns_gml("id"): self.get_gml_id(coverage.identifier)
            }
        )

    def encode_dataset_series_description(self, dataset_series):
        elements = []
        if dataset_series.footprint:
            elements.append(
                self.encode_bounded_by(dataset_series, None)
            )

        elements.append(EOWCS("DatasetSeriesId", dataset_series.identifier))

        if dataset_series.begin_time and dataset_series.end_time:
            elements.append(
                self.encode_time_period(
                    dataset_series.begin_time, dataset_series.end_time,
                    "%s_timeperiod" % dataset_series.identifier
                )
            )

        return EOWCS("DatasetSeriesDescription",
            *elements,
            **{ns_gml("id"): self.get_gml_id(dataset_series.identifier)}
        )

    def encode_dataset_series_descriptions(self, dataset_series_set):
        return EOWCS("DatasetSeriesDescriptions", *[
            self.encode_dataset_series_description(dataset_series)
            for dataset_series in dataset_series_set
        ])

    def encode_eo_coverage_set_description(self, dataset_series_set, coverages,
                                           number_matched=None,
                                           number_returned=None):
        if number_matched is None:
            number_matched = len(coverages) + len(dataset_series_set)
        if number_returned is None:
            number_returned = len(coverages) + len(dataset_series_set)

        root = EOWCS("EOCoverageSetDescription",
            numberMatched=str(number_matched),
            numberReturned=str(number_returned)
        )

        if coverages:
            root.append(self.encode_coverage_descriptions(coverages))
        if dataset_series_set:
            root.append(self.encode_dataset_series_descriptions(
                dataset_series_set
            ))

        return root

    def get_schema_locations(self):
        return {ns_eowcs.uri: ns_eowcs.schema_location}



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import tarfile

from eoxserver.core import Component, implements
from eoxserver.services.ows.wcs.interfaces import (
    PackageWriterInterface
)


gzip_mimes = ("application/gzip", "application/x-gzip")
bzip_mimes = ("application/bzip", "application/x-bzip")
mime_list = ("application/tar", "application/x-tar") + gzip_mimes + bzip_mimes


class TarPackageWriter(Component):
    """ Package writer for compressed and uncompressed tar files.
    """

    implements(PackageWriterInterface)

    def supports(self, format, params):
        return format.lower() in mime_list

    def create_package(self, filename, format, params):
        if format in gzip_mimes:
            mode = "w:gz"
        elif format in bzip_mimes:
            mode = "w:bz2"
        else:
            mode = "w"

        return tarfile.open(filename, mode)

    def cleanup(self, package):
        package.close()

    def add_to_package(self, package, file_obj, size, location):
        info = tarfile.TarInfo(location)
        info.size = size
        package.addfile(info, file_obj)

    def get_mime_type(self, package, format, params):
        return "application/x-compressed-tar"

    def get_file_extension(self, package, format, params):#
        if format in gzip_mimes:
            return ".tar.gz"

        elif format in bzip_mimes:
            return ".tar.bz2"

        return ".tar"



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.services.parameters import (
    VersionedParams, CapabilitiesRenderParams
)


class WCSParamsMixIn(object):
    def __iter__(self):
        for k, v in super(WCSParamsMixIn, self).__iter__():
            yield k, v
        
        yield ("service", "WCS")


class WCSCapabilitiesRenderParams(WCSParamsMixIn, CapabilitiesRenderParams):
    pass


class CoverageRenderParams(WCSParamsMixIn, VersionedParams):
    def __init__(self, coverage, version):
        super(CoverageRenderParams, self).__init__(version)
        self._coverage = coverage


    @property
    def coverage(self):
        return self._coverage


    @property
    def coverage_id(self):
        return self.coverage.identifier


    coverage_id_key_name = None

    def __iter__(self):
        for k, v in super(CoverageRenderParams, self).__iter__():
            yield k, v

        yield ("request", "GetCoverage")
        yield (self.coverage_id_key_name, self.coverage_id)


class CoverageDescriptionRenderParams(WCSParamsMixIn, VersionedParams):
    def __init__(self, coverages, version):
        super(CoverageDescriptionRenderParams, self).__init__(version)
        self._coverages = coverages

    @property
    def coverages(self):
        return self._coverages

    @property
    def coverage_ids(self):
        return [coverage.identifier for coverage in self.coverages]


    coverage_ids_key_name = None

    def __iter__(self):
        for k, v in super(CoverageDescriptionRenderParams, self).__iter__():
            yield k, v

        yield ("request", "DescribeCoverage")
        yield (self.coverage_ids_key_name, ",".join(self.coverage_ids))




# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import itertools

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.services import models as service_models
from eoxserver.resources.coverages.management.commands import CommandOutputMixIn


class Command(CommandOutputMixIn, BaseCommand):
    """ Command to manage service visibility for EOObjects (Collections, Mosaics,
        Products and Coverages).
    """
    def add_arguments(self, parser):
        """
        """

        service_choices = service_models.ServiceVisibility.SERVICE_CHOICES

        parser.add_argument(
            'identifiers', nargs='+', metavar='identifier',
            help='List of identifiers to change the visibility of.'
        )

        parser.add_argument(
            '--service', dest='services', action='append', default=[],
            help=(
                'Add a service to change the visibility of for the given objects. '
                'Possible values are: %s' % ', '.join(
                    '%s (%s)' % choice
                    for choice in service_choices
                )
            )
        )

        for key, name in service_choices:
            parser.add_argument(
                '--%s' % key, dest='services', action='append_const', const=key,
                help=(
                    'Add the service %s to change the visibility of for the given '
                    'objects.' % name
                )
            )

        parser.add_argument(
            '--all', dest='services', action='store_const',
            const=[choice[0] for choice in service_choices],
            help='Change the visibility for all available services.'
        )

        parser.add_argument(
            '--hide', dest='visibility', default=True, action='store_false',
            help='Set the visibility to False.'
        )

        parser.add_argument(
            '--show', dest='visibility', default=True, action='store_true',
            help='Set the visibility to True (the default).'
        )

    @transaction.atomic
    def handle(self, identifiers, services, visibility, *args, **kwargs):
        """
        """
        if not identifiers:
            raise CommandError('No identifiers specified.')

        if not services:
            raise CommandError('No services specified.')

        for identifier, service in itertools.product(identifiers, services):
            try:
                eo_object = models.EOObject.objects.get(identifier=identifier)
            except models.EOObject.DoesNotExist:
                raise CommandError('No such EOObject %s' % identifier)

            service_visibility, _ = service_models.ServiceVisibility.objects.get_or_create(
                eo_object=eo_object, service=service,
            )

            service_visibility.visibility = visibility
            service_visibility.full_clean()
            service_visibility.save()

        self.print_msg(
            'Successfully changed visibility of %d objects for the services: %s' % (
                len(identifiers), ', '.join(services)
            )
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.contrib.gis import admin
try:
    from django.core.urlresolvers import reverse, NoReverseMatch
except ImportError:
    from django.urls import reverse, NoReverseMatch
from django.utils.safestring import mark_safe
from django.conf import settings

from eoxserver.resources.coverages import models


# ==============================================================================
# Inline "Type" model admins
# ==============================================================================


class FieldTypeInline(admin.StackedInline):
    model = models.FieldType
    filter_horizontal = ['nil_values']
    extra = 0

    def get_queryset(self, *args, **kwargs):
        queryset = super(FieldTypeInline, self).get_queryset(*args, **kwargs)
        return queryset.order_by("index")


class MaskTypeInline(admin.TabularInline):
    model = models.MaskType
    extra = 0


class BrowseTypeInline(admin.StackedInline):
    model = models.BrowseType
    extra = 0

    fieldsets = (
        (None, {
            'fields': ('product_type', 'name')
        }),
        ("Red or grey band", {
            'classes': ('collapse', 'collapsed'),
            'fields': (
                'red_or_grey_expression', 'red_or_grey_nodata_value',
                ('red_or_grey_range_min', 'red_or_grey_range_max'),
            )
        }),
        ("Green band", {
            'classes': ('collapse', 'collapsed'),
            'fields': (
                'green_expression', 'green_nodata_value',
                ('green_range_min', 'green_range_max'),
            )
        }),
        ("Blue band", {
            'classes': ('collapse', 'collapsed'),
            'fields': (
                'blue_expression', 'blue_nodata_value',
                ('blue_range_min', 'blue_range_max'),
            )
        }),
        ("Alpha band", {
            'classes': ('collapse', 'collapsed'),
            'fields': (
                'alpha_expression', 'alpha_nodata_value',
                ('alpha_range_min', 'alpha_range_max'),
            )
        })
    )


# ==============================================================================
# Inline admins
# ==============================================================================

class MaskInline(admin.StackedInline):
    model = models.Mask
    extra = 0


class BrowseInline(admin.StackedInline):
    model = models.Browse
    extra = 0


class ProductDataItemInline(admin.StackedInline):
    model = models.ProductDataItem
    extra = 0


class MetaDataItemInline(admin.StackedInline):
    model = models.MetaDataItem
    extra = 0

    def download_link(self, obj):
        try:
            return mark_safe('<a href="{}">Download</a>'.format(
                reverse('metadata', kwargs=dict(
                        identifier=obj.eo_object.identifier,
                        semantic=dict(
                            models.MetaDataItem.SEMANTIC_CHOICES
                        )[obj.semantic]
                    )
                )
            ))
        except (NoReverseMatch, KeyError):
            return mark_safe('<i>Metadata URL not configured.</i>')
    download_link.short_description = 'Download Link'

    readonly_fields = ['download_link']


class ArrayDataItemInline(admin.StackedInline):
    model = models.ArrayDataItem
    extra = 0


class CoverageMetadataInline(admin.StackedInline):
    model = models.CoverageMetadata
    extra = 0


class ProductMetadataInline(admin.StackedInline):
    model = models.ProductMetadata
    extra = 0


class CollectionMetadataInline(admin.StackedInline):
    model = models.CollectionMetadata
    extra = 0


# ==============================================================================
# Abstract admins
# ==============================================================================

class EOObjectAdmin(admin.GeoModelAdmin):
    date_hierarchy = 'inserted'

    @property
    def wms_name(self):
        return getattr(settings, 'EOXS_ADMIN_WMS_NAME', 'EOX Maps')

    @property
    def wms_url(self):
        return getattr(settings, 'EOXS_ADMIN_WMS_URL', '//tiles.maps.eox.at/wms/')

    @property
    def wms_layer(self):
        return getattr(settings, 'EOXS_ADMIN_WMS_LAYER', 'terrain-light')

    @property
    def default_lon(self):
        return getattr(settings, 'EOXS_ADMIN_DEFAULT_LON', 16)

    @property
    def default_lat(self):
        return getattr(settings, 'EOXS_ADMIN_DEFAULT_LAT', 48)

    @property
    def default_zoom(self):
        return getattr(settings, 'EOXS_ADMIN_DEFAULT_ZOOM', 4)

# ==============================================================================
# "Type" model admins
# ==============================================================================


class CoverageTypeAdmin(admin.ModelAdmin):
    inlines = [FieldTypeInline]

admin.site.register(models.CoverageType, CoverageTypeAdmin)


class ProductTypeAdmin(admin.ModelAdmin):
    inlines = [BrowseTypeInline, MaskTypeInline]
    filter_horizontal = ['allowed_coverage_types']

admin.site.register(models.ProductType, ProductTypeAdmin)


class CollectionTypeAdmin(admin.ModelAdmin):
    filter_horizontal = ['allowed_product_types', 'allowed_coverage_types']

admin.site.register(models.CollectionType, CollectionTypeAdmin)


class MaskTypeAdmin(admin.ModelAdmin):
    pass

admin.site.register(models.MaskType, MaskTypeAdmin)


class GridAdmin(admin.ModelAdmin):
    pass

admin.site.register(models.Grid, GridAdmin)

# ==============================================================================
# Collection, Product and Coverage admins
# ==============================================================================


class CoverageAdmin(EOObjectAdmin):
    inlines = [CoverageMetadataInline, MetaDataItemInline, ArrayDataItemInline]

admin.site.register(models.Coverage, CoverageAdmin)


class ProductAdmin(EOObjectAdmin):
    inlines = [
        MaskInline, BrowseInline, ProductDataItemInline, MetaDataItemInline, ProductMetadataInline
    ]

admin.site.register(models.Product, ProductAdmin)


class MosaicAdmin(EOObjectAdmin):
    inlines = []

admin.site.register(models.Mosaic, MosaicAdmin)


class CollectionAdmin(EOObjectAdmin):
    inlines = [CollectionMetadataInline]

    actions = ['summary']

    # action to refresh the summary info on a collection
    def summary(self, request, queryset):
        for collection in queryset:
            models.collection_collect_metadata(
                collection, product_summary=True, coverage_summary=True
            )

    summary.short_description = (
        "Update the summary information for each collection"
    )


admin.site.register(models.Collection, CollectionAdmin)


class IndexHiddenAdmin(admin.ModelAdmin):
    """ Admin class that hides on the apps admin index page.
        """
    def get_model_perms(self, request):
        return {}

admin.site.register(models.OrbitNumber, IndexHiddenAdmin)
admin.site.register(models.Track, IndexHiddenAdmin)
admin.site.register(models.Frame, IndexHiddenAdmin)
admin.site.register(models.SwathIdentifier, IndexHiddenAdmin)
admin.site.register(models.ProductVersion, IndexHiddenAdmin)
admin.site.register(models.ProductQualityDegredationTag, IndexHiddenAdmin)
admin.site.register(models.ProcessorName, IndexHiddenAdmin)
admin.site.register(models.ProcessingCenter, IndexHiddenAdmin)
admin.site.register(models.SensorMode, IndexHiddenAdmin)
admin.site.register(models.ArchivingCenter, IndexHiddenAdmin)
admin.site.register(models.ProcessingMode, IndexHiddenAdmin)
admin.site.register(models.AcquisitionStation, IndexHiddenAdmin)
admin.site.register(models.AcquisitionSubType, IndexHiddenAdmin)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from os.path import join
from uuid import uuid4

from eoxserver.contrib import vsi, vrt, mapserver, gdal
from eoxserver.resources.coverages import models
from eoxserver.processing.gdal import reftools


class MultiFileConnector(object):
    """ Connects multiple files containing the various bands of the coverage
        with the given layer. A temporary VRT file is used as abstraction for
        the different band files.
    """

    def supports(self, coverage, data_items):
        # TODO: better checks
        return len(data_items) > 1

    def connect(self, coverage, data_items, layer, options):
        path = join("/vsimem", uuid4().hex)
        range_type = coverage.range_type

        # size_x, size_y = coverage.size[:2]
        # vrt_builder = vrt.VRTBuilder(size_x, size_y, vrt_filename=path)

        # for data_item in data_items:
        #     start = data_item.start_field
        #     end = data_item.end_field
        #     if end is None:
        #         dst_band_indices = range(start+1, start+2)
        #         src_band_indices = range(1, 2)
        #     else:
        #         dst_band_indices = range(start+1, end+2)
        #         src_band_indices = range(1, end-start+1)

        #     for src_index, dst_index in zip(src_band_indices, dst_band_indices):
        #         vrt_builder.add_band(range_type[dst_index-1].data_type)
        #         vrt_builder.add_simple_source(
        #             dst_index,
        #             data_item.path,
        #             src_index
        #         )

        # if coverage.grid.is_referenceable:
        #     vrt_builder.copy_gcps(gdal.OpenShared(data_items[0].path))
        #     layer.setMetaData("eoxs_ref_data", path)

        # layer.data = path

        # del vrt_builder

        # with vsi.open(path) as f:
        #     print f.read(100000)

        vrt.gdalbuildvrt(path, [
            location.path
            for location in coverage.arraydata_locations
        ], separate=True)

        layer.data = path

        #layer.clearProcessing()
        #layer.addProcessing("SCALE_1=1,4")
        #layer.addProcessing("BANDS=2")
        #layer.offsite = mapserver.colorObj(0,0,0)

        if coverage.grid.is_referenceable:
            vrt_path = join("/vsimem", uuid4().hex)
            reftools.create_rectified_vrt(path, vrt_path)
            layer.data = vrt_path
            layer.setMetaData("eoxs_ref_data", path)

            with vsi.open(vrt_path) as f:
                print (f.read(100000))

        """
        # TODO!!
        if layer.metadata.get("eoxs_wrap_dateline") == "true":
            e = wrap_extent_around_dateline(coverage.extent, coverage.srid)

            vrt_path = join("/vsimem", uuid4().hex)
            ds = gdal.Open(data)
            vrt_ds = create_simple_vrt(ds, vrt_path)
            size_x = ds.RasterXSize
            size_y = ds.RasterYSize

            dx = abs(e[0] - e[2]) / size_x
            dy = abs(e[1] - e[3]) / size_y

            vrt_ds.SetGeoTransform([e[0], dx, 0, e[3], 0, -dy])
            vrt_ds = None

            layer.data = vrt_path
        """

    def disconnect(self, coverage, data_items, layer, options):
        try:
            vsi.remove(layer.data)
        except:
            pass

        vrt_path = layer.metadata.get("eoxs_ref_data")
        if vrt_path:
            vsi.remove(vrt_path)
